import os
import sys
import time

import pytest
from client import SoraClient, SoraRole

from sora_sdk import (
    Sora,
    SoraVideoCodecImplementation,
    SoraVideoCodecPreference,
    SoraVideoCodecType,
    create_video_codec_preference_from_implementation,
    get_video_codec_capability,
)

pytestmark = pytest.mark.skipif(
    os.environ.get("OPENH264_PATH") is None,
    reason="OpenH264 のときだけ実行する",
)


def test_capability(settings):
    capability = get_video_codec_capability(openh264=settings.openh264_path)
    has_internal = False
    has_openh264 = False
    for engine in capability.engines:
        if engine.name == SoraVideoCodecImplementation.INTERNAL:
            has_internal = True
        if engine.name == SoraVideoCodecImplementation.CISCO_OPENH264:
            has_openh264 = True
    assert has_internal and has_openh264


def test_preference(settings):
    capability = get_video_codec_capability(openh264=settings.openh264_path)
    preference = create_video_codec_preference_from_implementation(
        capability, SoraVideoCodecImplementation.CISCO_OPENH264
    )
    assert preference.has_implementation(SoraVideoCodecImplementation.CISCO_OPENH264)


def test_openh264_get_codec_capability(settings):
    capability = get_video_codec_capability(openh264=settings.openh264_path)

    cisco_openh264_available = False

    for e in capability.engines:
        if e.name == SoraVideoCodecImplementation.CISCO_OPENH264:
            cisco_openh264_available = True
    assert cisco_openh264_available is True

    codecs = []
    for e in capability.engines:
        # macOS では INTERNAL の Video Toolbox が利用されるので OpenH264 は採用しないようにする
        if sys.platform == "darwin":
            if e.name == SoraVideoCodecImplementation.CISCO_OPENH264:
                continue

        if e.name == SoraVideoCodecImplementation.CISCO_OPENH264:
            for c in e.codecs:
                if c.decoder or c.encoder:
                    # encoder/decoder どちらかが true であれば採用する
                    if c.decoder or c.encoder:
                        codecs.append(
                            SoraVideoCodecPreference.Codec(
                                type=c.type,
                                decoder=e.name,
                                encoder=e.name,
                                parameters=SoraVideoCodecPreference.Parameters(),
                            )
                        )

    # エラーにならないことを確認する
    Sora(
        video_codec_preference=SoraVideoCodecPreference(
            codecs=codecs,
        ),
        openh264=settings.openh264_path,
    )


def test_openh264_video_codec_preference(settings):
    Sora(
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.VP8,
                    decoder=SoraVideoCodecImplementation.INTERNAL,
                    encoder=SoraVideoCodecImplementation.INTERNAL,
                ),
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.VP9,
                    decoder=SoraVideoCodecImplementation.INTERNAL,
                    encoder=SoraVideoCodecImplementation.INTERNAL,
                ),
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.AV1,
                    decoder=SoraVideoCodecImplementation.INTERNAL,
                    encoder=SoraVideoCodecImplementation.INTERNAL,
                ),
                # H.264 だけは OpenH264 を利用するようにする
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    decoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                    encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                ),
            ],
        ),
        openh264=settings.openh264_path,
    )


def test_openh264_sendonly_recvonly(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type="H264",
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                )
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    decoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                )
            ]
        ),
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == "video/H264"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "OpenH264"
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    assert recvonly_codec_stats["mimeType"] == "video/H264"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "OpenH264"
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0



import time

import pytest
from client import SoraClient, SoraRole


def test_websocket_signaling_only_type_switched(settings):
    """
    - WebSocket シグナリングのみ
    - type: switched 送られてこない
    """

    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=False,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        assert conn.switched is False
        assert conn.ws_close_code is None
        assert conn.ws_close_reason is None


def test_hybrid_signaling_type_switched(settings):
    """
    - WebSocket シグナリング + DataChannel シグナリング
    - type: switched 送られてくる
    """
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        assert conn.switched is True
        assert conn.ws_close_code is None
        assert conn.ws_close_reason is None


def test_datachannel_signaling_only_type_switched(settings):
    """
    - DataChannel シグナリングのみ
    - type: switched 送られてくる
    - Python SDK は WebSocket を自分で切断する
    """
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=True,
    ) as conn:
        time.sleep(3)

        assert conn.switched is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "SELF-CLOSED"


@pytest.mark.skip(reason="Sora がまだ対応していない")
def test_disconnect_before_switched(settings):
    # switched 前に type: disconnect を送りつける
    # ignore_disconnect_websocket は true

    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=True,
    ) as conn:
        conn.disconnect()

        if not conn.switched:
            assert conn.ws_close_code == 1000
            assert conn.ws_close_reason == "TYPE-DISCONNECT"



from collections.abc import Callable, Sequence
import enum
from typing import Annotated, overload

import numpy
from numpy.typing import NDArray


class SoraSignalingErrorCode(enum.IntEnum):
    CLOSE_SUCCEEDED = 0

    CLOSE_FAILED = 1

    INTERNAL_ERROR = 2

    INVALID_PARAMETER = 3

    WEBSOCKET_HANDSHAKE_FAILED = 4

    WEBSOCKET_ONCLOSE = 5

    WEBSOCKET_ONERROR = 6

    PEER_CONNECTION_STATE_FAILED = 7

    ICE_FAILED = 8

class SoraSignalingType(enum.IntEnum):
    WEBSOCKET = 0

    DATACHANNEL = 1

class SoraDegradationPreference(enum.IntEnum):
    DISABLED = 0

    BALANCED = 3

    MAINTAIN_FRAMERATE = 1

    MAINTAIN_RESOLUTION = 2

class SoraSignalingDirection(enum.IntEnum):
    SENT = 0

    RECEIVED = 1

class SoraTrackState(enum.IntEnum):
    LIVE = 0

    ENDED = 1

class SoraLoggingSeverity(enum.IntEnum):
    VERBOSE = 0

    INFO = 1

    WARNING = 2

    ERROR = 3

    NONE = 4

def enable_libwebrtc_log(arg: SoraLoggingSeverity, /) -> None: ...

def rtc_log(arg0: SoraLoggingSeverity, arg1: str, /) -> None: ...

class SoraTrackInterface:
    @property
    def kind(self) -> str: ...

    @property
    def id(self) -> str: ...

    @property
    def enabled(self) -> bool: ...

    @property
    def state(self) -> SoraTrackState: ...

    def set_enabled(self, enable: bool) -> bool: ...

class SoraMediaTrack(SoraTrackInterface):
    @property
    def stream_id(self) -> str: ...

    def set_frame_transformer(self, arg: SoraFrameTransformer, /) -> None: ...

class SoraAudioSource(SoraTrackInterface):
    @overload
    def on_data(self, data: int, samples_per_channel: int, timestamp: float) -> None: ...

    @overload
    def on_data(self, data: int, samples_per_channel: int) -> None: ...

    @overload
    def on_data(self, ndarray: Annotated[NDArray[numpy.int16], dict(shape=(None, None), order='C', device='cpu')], timestamp: float) -> None: ...

    @overload
    def on_data(self, ndarray: Annotated[NDArray[numpy.int16], dict(shape=(None, None), order='C', device='cpu')]) -> None: ...

class SoraVideoSource(SoraTrackInterface):
    @overload
    def on_captured(self, ndarray: Annotated[NDArray[numpy.uint8], dict(shape=(None, None, 3), order='C', device='cpu')]) -> None: ...

    @overload
    def on_captured(self, ndarray: Annotated[NDArray[numpy.uint8], dict(shape=(None, None, 3), order='C', device='cpu')], timestamp: float) -> None: ...

    @overload
    def on_captured(self, ndarray: Annotated[NDArray[numpy.uint8], dict(shape=(None, None, 3), order='C', device='cpu')], timestamp_us: int) -> None: ...

class SoraAudioSinkImpl:
    def __init__(self, track: SoraTrackInterface, output_frequency: int = -1, output_channels: int = 0) -> None: ...

    def __del__(self) -> None: ...

    def read(self, frames: int = 0, timeout: float = 1) -> tuple: ...

    @property
    def on_data(self) -> Callable[[Annotated[NDArray[numpy.int16], dict(shape=(None, None))]], None]: ...

    @on_data.setter
    def on_data(self, arg: Callable[[Annotated[NDArray[numpy.int16], dict(shape=(None, None))]], None], /) -> None: ...

    @property
    def on_format(self) -> Callable[[int, int], None]: ...

    @on_format.setter
    def on_format(self, arg: Callable[[int, int], None], /) -> None: ...

class SoraAudioFrame:
    def __getstate__(self) -> tuple[list[int], int, int, int, int | None]: ...

    def __setstate__(self, arg: tuple[Sequence[int], int, int, int, int | None], /) -> None: ...

    @property
    def samples_per_channel(self) -> int: ...

    @property
    def num_channels(self) -> int: ...

    @property
    def sample_rate_hz(self) -> int: ...

    @property
    def absolute_capture_timestamp_ms(self) -> int | None: ...

    def data(self) -> Annotated[NDArray[numpy.int16], dict(shape=(None, None))]: ...

class SoraAudioStreamSinkImpl:
    def __init__(self, track: SoraTrackInterface, output_frequency: int = -1, output_channels: int = 0) -> None: ...

    def __del__(self) -> None: ...

    @property
    def on_frame(self) -> Callable[[SoraAudioFrame], None]: ...

    @on_frame.setter
    def on_frame(self, arg: Callable[[SoraAudioFrame], None], /) -> None: ...

class SoraVAD:
    def __init__(self) -> None: ...

    def analyze(self, frame: SoraAudioFrame) -> float: ...

class SoraVideoFrame:
    def data(self) -> Annotated[NDArray[numpy.uint8], dict(shape=(None, None, 3))]: ...

class SoraVideoSinkImpl:
    def __init__(self, arg: SoraTrackInterface, /) -> None: ...

    def __del__(self) -> None: ...

    @property
    def on_frame(self) -> Callable[[SoraVideoFrame], None]: ...

    @on_frame.setter
    def on_frame(self, arg: Callable[[SoraVideoFrame], None], /) -> None: ...

class SoraConnection:
    def connect(self) -> None: ...

    def disconnect(self) -> None: ...

    def send_data_channel(self, label: str, data: bytes) -> bool: ...

    def get_stats(self) -> str: ...

    @property
    def on_set_offer(self) -> Callable[[str], None]: ...

    @on_set_offer.setter
    def on_set_offer(self, arg: Callable[[str], None], /) -> None: ...

    @property
    def on_ws_close(self) -> Callable[[int, str], None]: ...

    @on_ws_close.setter
    def on_ws_close(self, arg: Callable[[int, str], None], /) -> None: ...

    @property
    def on_disconnect(self) -> Callable[[SoraSignalingErrorCode, str], None]: ...

    @on_disconnect.setter
    def on_disconnect(self, arg: Callable[[SoraSignalingErrorCode, str], None], /) -> None: ...

    @property
    def on_signaling_message(self) -> Callable[[SoraSignalingType, SoraSignalingDirection, str], None]: ...

    @on_signaling_message.setter
    def on_signaling_message(self, arg: Callable[[SoraSignalingType, SoraSignalingDirection, str], None], /) -> None: ...

    @property
    def on_notify(self) -> Callable[[str], None]: ...

    @on_notify.setter
    def on_notify(self, arg: Callable[[str], None], /) -> None: ...

    @property
    def on_push(self) -> Callable[[str], None]: ...

    @on_push.setter
    def on_push(self, arg: Callable[[str], None], /) -> None: ...

    @property
    def on_message(self) -> Callable[[str, bytes], None]: ...

    @on_message.setter
    def on_message(self, arg: Callable[[str, bytes], None], /) -> None: ...

    @property
    def on_rpc(self) -> Callable[[bytes], None]: ...

    @on_rpc.setter
    def on_rpc(self, arg: Callable[[bytes], None], /) -> None: ...

    @property
    def on_switched(self) -> Callable[[str], None]: ...

    @on_switched.setter
    def on_switched(self, arg: Callable[[str], None], /) -> None: ...

    @property
    def on_track(self) -> Callable[[SoraMediaTrack], None]: ...

    @on_track.setter
    def on_track(self, arg: Callable[[SoraMediaTrack], None], /) -> None: ...

    @property
    def on_data_channel(self) -> Callable[[str], None]: ...

    @on_data_channel.setter
    def on_data_channel(self, arg: Callable[[str], None], /) -> None: ...

class SoraTransformableFrameDirection(enum.IntEnum):
    UNKNOWN = 0

    RECEIVER = 1

    SENDER = 2

class SoraTransformableFrame:
    def get_data(self) -> Annotated[NDArray[numpy.uint8], dict(shape=(None,), writable=False)]: ...

    def set_data(self, arg: Annotated[NDArray[numpy.uint8], dict(shape=(None,), order='C', device='cpu', writable=False)], /) -> None: ...

    @property
    def payload_type(self) -> int: ...

    @property
    def ssrc(self) -> int: ...

    @property
    def rtp_timestamp(self) -> int: ...

    @rtp_timestamp.setter
    def rtp_timestamp(self, arg: int, /) -> None: ...

    @property
    def direction(self) -> SoraTransformableFrameDirection: ...

    @property
    def mine_type(self) -> str: ...

class SoraTransformableAudioFrameType(enum.IntEnum):
    EMPTY = 0

    SPEECH = 1

    CN = 2

class SoraTransformableAudioFrame(SoraTransformableFrame):
    @property
    def contributing_sources(self) -> Annotated[NDArray[numpy.uint32], dict(shape=(None,), writable=False)]: ...

    @property
    def sequence_number(self) -> int | None: ...

    @property
    def absolute_capture_timestamp(self) -> int | None: ...

    @property
    def type(self) -> SoraTransformableAudioFrameType: ...

    @property
    def audio_level(self) -> int | None: ...

    @property
    def receive_time(self) -> int | None: ...

class SoraTransformableVideoFrame(SoraTransformableFrame):
    @property
    def is_key_frame(self) -> bool: ...

    @property
    def frame_id(self) -> int | None: ...

    @property
    def frame_dependencies(self) -> Annotated[NDArray[numpy.int64], dict(shape=(None,), writable=False)]: ...

    @property
    def width(self) -> int: ...

    @property
    def height(self) -> int: ...

    @property
    def spatial_index(self) -> int: ...

    @property
    def temporal_index(self) -> int: ...

    @property
    def contributing_sources(self) -> Annotated[NDArray[numpy.uint32], dict(shape=(None,), writable=False)]: ...

class SoraFrameTransformer:
    def enqueue(self, arg: SoraTransformableFrame, /) -> None: ...

    def start_short_circuiting(self) -> None: ...

class SoraAudioFrameTransformer(SoraFrameTransformer):
    def __init__(self) -> None: ...

    def __del__(self) -> None: ...

    @property
    def on_transform(self) -> Callable[[SoraTransformableAudioFrame], None]: ...

    @on_transform.setter
    def on_transform(self, arg: Callable[[SoraTransformableAudioFrame], None], /) -> None: ...

class SoraVideoFrameTransformer(SoraFrameTransformer):
    def __init__(self) -> None: ...

    def __del__(self) -> None: ...

    @property
    def on_transform(self) -> Callable[[SoraTransformableVideoFrame], None]: ...

    @on_transform.setter
    def on_transform(self, arg: Callable[[SoraTransformableVideoFrame], None], /) -> None: ...

class SoraVideoCodecImplementation(enum.IntEnum):
    INTERNAL = 0

    CISCO_OPENH264 = 1

    INTEL_VPL = 2

    NVIDIA_VIDEO_CODEC_SDK = 3

    AMD_AMF = 4

    RASPI_V4L2M2M = 5

class SoraVideoCodecType(enum.IntEnum):
    VP8 = 1

    VP9 = 2

    H264 = 4

    H265 = 5

    AV1 = 3

class SoraVideoCodecCapability:
    @property
    def engines(self) -> list[SoraVideoCodecCapability.Engine]: ...

    def to_json(self) -> object: ...

    class Parameters:
        @property
        def version(self) -> str | None: ...

        @property
        def openh264_path(self) -> str | None: ...

        @property
        def vpl_impl(self) -> str | None: ...

        @property
        def vpl_impl_value(self) -> int | None: ...

        @property
        def nvcodec_gpu_device_name(self) -> str | None: ...

        @property
        def amf_runtime_version(self) -> str | None: ...

        @property
        def amf_embedded_version(self) -> str | None: ...

    class Codec:
        @property
        def type(self) -> SoraVideoCodecType: ...

        @property
        def encoder(self) -> bool: ...

        @property
        def decoder(self) -> bool: ...

        @property
        def parameters(self) -> SoraVideoCodecCapability.Parameters: ...

    class Engine:
        @property
        def name(self) -> SoraVideoCodecImplementation: ...

        @property
        def codecs(self) -> list[SoraVideoCodecCapability.Codec]: ...

        @property
        def parameters(self) -> SoraVideoCodecCapability.Parameters: ...

def get_video_codec_capability(openh264: str | None = None) -> SoraVideoCodecCapability: ...

class SoraVideoCodecPreference:
    def __init__(self, codecs: Sequence[SoraVideoCodecPreference.Codec] = []) -> None: ...

    @property
    def codecs(self) -> list[SoraVideoCodecPreference.Codec]: ...

    @codecs.setter
    def codecs(self, arg: Sequence[SoraVideoCodecPreference.Codec], /) -> None: ...

    def to_json(self) -> object: ...

    def find(self, arg: SoraVideoCodecType, /) -> int | None: ...

    def get_or_add(self, arg: SoraVideoCodecType, /) -> int: ...

    def has_implementation(self, arg: SoraVideoCodecImplementation, /) -> bool: ...

    def merge(self, arg: SoraVideoCodecPreference, /) -> None: ...

    class Parameters:
        def __init__(self) -> None: ...

    class Codec:
        @overload
        def __init__(self) -> None: ...

        @overload
        def __init__(self, type: SoraVideoCodecType, encoder: SoraVideoCodecImplementation | None = None, decoder: SoraVideoCodecImplementation | None = None, parameters: SoraVideoCodecPreference.Parameters | None = None) -> None: ...

        @property
        def type(self) -> SoraVideoCodecType: ...

        @type.setter
        def type(self, arg: SoraVideoCodecType, /) -> None: ...

        @property
        def encoder(self) -> SoraVideoCodecImplementation | None: ...

        @encoder.setter
        def encoder(self, arg: SoraVideoCodecImplementation, /) -> None: ...

        @property
        def decoder(self) -> SoraVideoCodecImplementation | None: ...

        @decoder.setter
        def decoder(self, arg: SoraVideoCodecImplementation, /) -> None: ...

        @property
        def parameters(self) -> SoraVideoCodecPreference.Parameters: ...

        @parameters.setter
        def parameters(self, arg: SoraVideoCodecPreference.Parameters, /) -> None: ...

def create_video_codec_preference_from_implementation(arg0: SoraVideoCodecCapability, arg1: SoraVideoCodecImplementation, /) -> SoraVideoCodecPreference: ...

class Sora:
    def __init__(self, openh264: str | None = None, video_codec_preference: SoraVideoCodecPreference | None = None, force_i420_conversion: bool | None = None) -> None: ...

    def create_connection(self, signaling_urls: list[str], role: str, channel_id: str, client_id: Optional[str] = None, bundle_id: Optional[str] = None, metadata: Optional[dict] = None, signaling_notify_metadata: Optional[dict] = None, audio_source: Optional[SoraTrackInterface] = None, video_source: Optional[SoraTrackInterface] = None, audio_frame_transformer: Optional[SoraAudioFrameTransformer] = None, video_frame_transformer: Optional[SoraVideoFrameTransformer] = None, audio: Optional[bool] = None, video: Optional[bool] = None, audio_codec_type: Optional[str] = None, video_codec_type: Optional[str] = None, video_bit_rate: Optional[int] = None, audio_bit_rate: Optional[int] = None, video_vp9_params: Optional[dict] = None, video_av1_params: Optional[dict] = None, video_h264_params: Optional[dict] = None, audio_opus_params: Optional[dict] = None, simulcast: Optional[bool] = None, spotlight: Optional[bool] = None, spotlight_number: Optional[int] = None, simulcast_rid: Optional[str] = None, spotlight_focus_rid: Optional[str] = None, spotlight_unfocus_rid: Optional[str] = None, forwarding_filter: Optional[dict] = None, forwarding_filters: Optional[list[dict]] = None, data_channels: Optional[list[dict]] = None, data_channel_signaling: Optional[bool] = None, ignore_disconnect_websocket: Optional[bool] = None, data_channel_signaling_timeout: Optional[int] = None, disconnect_wait_timeout: Optional[int] = None, websocket_close_timeout: Optional[int] = None, websocket_connection_timeout: Optional[int] = None, audio_streaming_language_code: Optional[str] = None, insecure: Optional[bool] = None, client_cert: Optional[bytes] = None, client_key: Optional[bytes] = None, ca_cert: Optional[bytes] = None, proxy_url: Optional[str] = None, proxy_username: Optional[str] = None, proxy_password: Optional[str] = None, proxy_agent: Optional[str] = None, degradation_preference: Optional[SoraDegradationPreference] = None, user_agent: Optional[str] = None) -> SoraConnection: ...

    def create_audio_source(self, channels: int, sample_rate: int) -> SoraAudioSource: ...

    def create_video_source(self) -> SoraVideoSource: ...

    def create_libcamera_source(self, width: int, height: int, fps: int, native_frame_output: bool, controls: Optional[list[tuple[str, str]]] = None) -> SoraTrackInterface: ...



import time

import pytest
from client import SoraClient, SoraRole


def test_sendonly_recvonly_audio(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=False,
    )
    sendonly.connect(fake_audio=True)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == "audio/opus"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    # audio には encoderImplementation が無い
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    assert recvonly_codec_stats["mimeType"] == "audio/opus"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    # audio には decoderImplementation が無い
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0


@pytest.mark.parametrize(
    "video_codec_type, encoder_implementation, decoder_implementation",
    [
        ("VP8", "libvpx", "libvpx"),
        ("VP9", "libvpx", "libvpx"),
        ("AV1", "libaom", "dav1d"),
    ],
)
def test_sendonly_recvonly_video(
    settings, video_codec_type, encoder_implementation, decoder_implementation
):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
    )
    sendonly.connect(fake_video=True)

    time.sleep(5)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == encoder_implementation
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0
    assert outbound_rtp_stats["keyFramesEncoded"] > 0
    assert outbound_rtp_stats["pliCount"] > 0

    print("keyFramesEncoded:", outbound_rtp_stats["keyFramesEncoded"])
    print("pliCount:", outbound_rtp_stats["pliCount"])

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    assert recvonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # inbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    assert inbound_rtp_stats["decoderImplementation"] == decoder_implementation
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0
    assert inbound_rtp_stats["keyFramesDecoded"] > 0

    print("keyFrameDecoded:", inbound_rtp_stats["keyFramesDecoded"])



import time

from client import SoraClient, SoraRole

from sora_sdk import (
    SoraVideoCodecImplementation,
    SoraVideoCodecPreference,
    SoraVideoCodecType,
)

# def test_preference_invalid(settings):
#     # opneh264 のパスを設定してないのに H264 エンコーダを使おうとしているので Sora 生成時にエラーになる
#     with pytest.raises(RuntimeError):
#         SoraClient(
#             settings,
#             SoraRole.SENDONLY,
#             audio=False,
#             video=True,
#             video_codec_type="H264",
#             video_codec_preference=SoraVideoCodecPreference(
#                 codecs=[
#                     SoraVideoCodecPreference.Codec(
#                         type=SoraVideoCodecType.H264,
#                         encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
#                     )
#                 ]
#             ),
#         )


def test_preference_vp8(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type="VP8",
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.VP8, encoder=SoraVideoCodecImplementation.INTERNAL
                )
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(5)

    sendonly.disconnect()



import json
import time
from threading import Event
from typing import Any

from client import SoraClient, SoraRole
from conftest import Settings

from sora_sdk import (
    Sora,
    SoraAudioFrame,
    SoraAudioStreamSink,
    SoraMediaTrack,
    SoraVAD,
)


class VAD:
    def __init__(
        self,
        settings: Settings,
        metadata: dict[str, Any] | None = None,
        jwt_private_claims: dict[str, Any] | None = None,
    ):
        self._signaling_urls: list[str] = settings.signaling_urls
        self._channel_id: str = settings.channel_id

        if jwt_private_claims is not None:
            access_token = settings.access_token(**jwt_private_claims)
        else:
            access_token = settings.access_token()

        # secret が設定されていない場合は access_token が存在しない
        if access_token is not None:
            if metadata is not None:
                # metadata が設定されている場合は access_token を追加する
                metadata.update({"access_token": access_token})
            else:
                # metadata が設定されていない場合は access_token のみを metadata に設定する
                metadata = {"access_token": access_token}

        self._vad = SoraVAD()

        self._connection_id: str

        # 接続した
        self._connected: Event = Event()
        # 終了
        self._disconnected = Event()

        self._audio_output_frequency: int = 24000
        self._audio_output_channels: int = 1

        self._sora = Sora()

        self._connection = self._sora.create_connection(
            signaling_urls=self._signaling_urls,
            role="recvonly",
            channel_id=self._channel_id,
            metadata=metadata,
            audio=True,
            video=False,
        )

        self._connection.on_set_offer = self._on_set_offer
        self._connection.on_notify = self._on_notify
        self._connection.on_disconnect = self._on_disconnect

        self._connection.on_track = self._on_track

    def connect(self):
        self._connection.connect()

        # _connected が set されるまで 30 秒待つ
        assert self._connected.wait(30)

        return self

    def disconnect(self):
        self._connection.disconnect()

    def get_stats(self):
        raw_stats = self._connection.get_stats()
        stats = json.loads(raw_stats)
        return stats

    def _on_set_offer(self, raw_offer):
        offer = json.loads(raw_offer)
        if offer["type"] == "offer":
            self._connection_id = offer["connection_id"]
            print(f"Received 'Offer': connection_id={self._connection_id}")

    def _on_notify(self, raw_message):
        message = json.loads(raw_message)
        if (
            message["type"] == "notify"
            and message["event_type"] == "connection.created"
            and message["connection_id"] == self._connection_id
        ):
            print(f"Connected Sora: connection_id={self._connection_id}")
            self._connected.set()

    def _on_disconnect(self, error_code, message):
        print(f"Disconnected Sora: error_code='{error_code}' message='{message}'")
        self._disconnected.set()
        self._connected.clear()

    def _on_frame(self, frame: SoraAudioFrame):
        # frame が音声である確率を求める
        voice_probability = self._vad.analyze(frame)
        if voice_probability > 0.95:  # 0.95 は libwebrtc の判定値
            print(f"Voice! voice_probability={voice_probability}")
        else:
            pass

    def _on_track(self, track: SoraMediaTrack):
        if track.kind == "audio":
            # SoraAudioStreamSink
            self._audio_stream_sink = SoraAudioStreamSink(
                track, self._audio_output_frequency, self._audio_output_channels
            )
            self._audio_stream_sink.on_frame = self._on_frame


def test_vad(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=False,
    )
    sendonly.connect(fake_audio=True)

    vad = VAD(settings)
    vad.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    vad_stats = vad.get_stats()

    sendonly.disconnect()
    vad.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == "audio/opus"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    # audio には encoderImplementation が無い
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    vad_codec_stats = next(s for s in vad_stats if s.get("type") == "codec")
    assert vad_codec_stats["mimeType"] == "audio/opus"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in vad_stats if s.get("type") == "inbound-rtp")
    # audio には decoderImplementation が無い
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0



import time

from client import SoraClient, SoraRole

# opus params のテストは test_signaling.py にある

# https://tex2e.github.io/rfc-translater/html/rfc7587.html


def test_sendonly_audio_opus_params_16khz_mono(settings):
    """
    SDP では 48000/2 だが、Opus の設定で 16000/1 を配信してみる
    """
    with SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        audio_codec_type="OPUS",
        audio_opus_params={
            #
            "maxplaybackrate": 16000,
            # 受信のみに有効
            "stereo": False,
            # 送信のみに有効
            "sprop_stereo": False,
        },
        video=False,
    ) as sendonly:
        time.sleep(5)

        assert sendonly.connect_message is not None
        assert "audio" in sendonly.connect_message
        assert "opus_params" in sendonly.connect_message["audio"]
        assert sendonly.connect_message["audio"]["opus_params"]["stereo"] is False
        assert sendonly.connect_message["audio"]["opus_params"]["maxplaybackrate"] == 16000
        assert sendonly.connect_message["audio"]["opus_params"]["sprop_stereo"] is False

        assert sendonly.offer_message is not None
        assert "sdp" in sendonly.offer_message

        assert "opus/48000/2" in sendonly.offer_message["sdp"]

        assert "maxplaybackrate=16000" in sendonly.offer_message["sdp"]
        assert "stereo=0" in sendonly.offer_message["sdp"]
        assert "sprop-stereo=0" in sendonly.offer_message["sdp"]

        sendonly_stats = sendonly.get_stats()

        sendonly.disconnect()

        # codec が無かったら StopIteration 例外が上がる
        sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
        assert sendonly_codec_stats["mimeType"] == "audio/opus"

        # outbound-rtp が無かったら StopIteration 例外が上がる
        outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
        assert outbound_rtp_stats["bytesSent"] > 0
        assert outbound_rtp_stats["packetsSent"] > 0



import time

from client import SoraClient, SoraRole


def test_sendonly_audio_opus_params_none(settings):
    with SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        audio_codec_type="OPUS",
        video=False,
    ) as sendonly:
        time.sleep(5)

        assert sendonly.connect_message is not None
        assert "audio" in sendonly.connect_message
        assert "opus_params" not in sendonly.connect_message["audio"]

        assert sendonly.offer_message is not None
        assert "sdp" in sendonly.offer_message
        ## usedtx=1 がない事を確認する
        assert "usedtx=1" not in sendonly.offer_message["sdp"]

        sendonly_stats = sendonly.get_stats()

        sendonly.disconnect()

        # codec が無かったら StopIteration 例外が上がる
        sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
        assert sendonly_codec_stats["mimeType"] == "audio/opus"

        # outbound-rtp が無かったら StopIteration 例外が上がる
        outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
        assert outbound_rtp_stats["bytesSent"] > 0
        assert outbound_rtp_stats["packetsSent"] > 0


def test_sendonly_audio_opus_params_usedtx_true(settings):
    with SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        audio_codec_type="OPUS",
        audio_opus_params={
            "usedtx": True,
        },
        video=False,
    ) as sendonly:
        time.sleep(5)

        assert sendonly.connect_message is not None
        assert "audio" in sendonly.connect_message
        assert "opus_params" in sendonly.connect_message["audio"]
        assert "usedtx" in sendonly.connect_message["audio"]["opus_params"]
        assert sendonly.connect_message["audio"]["opus_params"]["usedtx"] is True

        assert sendonly.offer_message is not None
        assert "sdp" in sendonly.offer_message
        ## usedtx=1 がある事を確認する
        assert "usedtx=1" in sendonly.offer_message["sdp"]

        sendonly_stats = sendonly.get_stats()

        sendonly.disconnect()

        # codec が無かったら StopIteration 例外が上がる
        sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
        assert sendonly_codec_stats["mimeType"] == "audio/opus"

        # outbound-rtp が無かったら StopIteration 例外が上がる
        outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
        assert outbound_rtp_stats["bytesSent"] > 0
        assert outbound_rtp_stats["packetsSent"] > 0



import time

from client import SoraClient, SoraRole


def test_messaging_header(settings):
    messaging_label = "#test"

    messaging_sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        data_channel_signaling=True,
        data_channels=[{"label": messaging_label, "direction": "sendonly"}],
    )

    messaging_recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
        data_channel_signaling=True,
        data_channels=[
            {
                "label": messaging_label,
                "direction": "recvonly",
                "header": [{"type": "sender_connection_id"}],
            }
        ],
    )

    # Sora に接続する
    messaging_sendonly.connect()
    messaging_recvonly.connect()

    time.sleep(3)

    # data_channel_signaling: true だし
    # switched のテストはここでやる
    assert messaging_sendonly.switched
    assert messaging_recvonly.switched

    message1 = "spam".encode("utf-8")
    message2 = "はむ".encode("utf-8")

    messaging_sendonly.send_message(messaging_label, message1)
    messaging_sendonly.send_message(messaging_label, message2)

    time.sleep(3)

    # 26 は sender_connection_id の長さ
    assert messaging_recvonly.recv_message(messaging_label)[26:] == message1
    assert messaging_recvonly.recv_message(messaging_label)[26:] == message2

    messaging_sendonly_stats = messaging_sendonly.get_stats()
    messaging_recvonly_stats = messaging_recvonly.get_stats()

    messaging_sendonly.disconnect()
    messaging_recvonly.disconnect()

    sendonly_data_channel_stats = next(
        s
        for s in messaging_sendonly_stats
        if s.get("type") == "data-channel" and s.get("label") == messaging_label
    )
    print(sendonly_data_channel_stats)
    assert sendonly_data_channel_stats["state"] == "open"
    assert sendonly_data_channel_stats["messagesSent"] == 2
    assert sendonly_data_channel_stats["bytesSent"] == (len(message1) + len(message2))

    recvonly_data_channel_stats = next(
        s
        for s in messaging_recvonly_stats
        if s.get("type") == "data-channel" and s.get("label") == messaging_label
    )
    assert recvonly_data_channel_stats["state"] == "open"
    assert recvonly_data_channel_stats["messagesReceived"] == 2
    # 26 は sender_connection_id の長さ x 2
    assert recvonly_data_channel_stats["bytesReceived"] == (26 + len(message1) + 26 + len(message2))



import os
import sys
import time

import pytest
from client import SoraClient, SoraRole
from simulcast import default_video_bit_rate, expect_target_bitrate


@pytest.mark.skipif(
    os.getenv("CI") == "true" and sys.platform == "darwin", reason="darwin では実行しない"
)
@pytest.mark.parametrize(
    (
        "video_codec_type",
        "encoder_implementation",
        "video_width",
        "video_height",
        "simulcast_count",
    ),
    [
        # 1080p
        ("VP8", "libvpx", 1920, 1080, 3),
        ("VP9", "libvpx", 1920, 1080, 3),
        ("AV1", "libaom", 1920, 1080, 3),
        # 720p
        ("VP8", "libvpx", 1280, 720, 3),
        ("VP9", "libvpx", 1280, 720, 3),
        ("AV1", "libaom", 1280, 720, 3),
        # 540p
        ("VP8", "libvpx", 960, 540, 3),
        ("VP9", "libvpx", 960, 540, 3),
        ("AV1", "libaom", 960, 540, 3),
        # simulcast count が 2 と 1 の場合解像度が 1/4 と 1/8 になってテストが通らなくなる
        # # 360p
        # ("VP8", "libvpx", 640, 360, 2),
        # ("VP9", "libvpx", 640, 360, 2),
        # ("AV1", "libaom", 640, 360, 2),
        # # 270p
        # ("VP8", "libvpx", 480, 270, 2),
        # ("VP9", "libvpx", 480, 270, 2),
        # ("AV1", "libaom", 480, 270, 2),
        # # 180p
        # ("VP8", "libvpx", 320, 180, 1),
        # ("VP9", "libvpx", 320, 180, 1),
        # ("AV1", "libaom", 320, 180, 1),
        # # 135p
        # ("VP9", "libvpx", 240, 135, 1),
    ],
)
def test_simulcast(
    settings,
    video_codec_type,
    encoder_implementation,
    video_width,
    video_height,
    simulcast_count,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;r2" in sendonly.answer_message["sdp"]

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        # simulcast_count が 2 の場合、rid r2 の bytesSent/packetsSent は 0 or 1 になる
        # simulcast_count が 1 の場合、rid r2 と r1 の bytesSent/packetsSent は 0 or 1 になる
        if i < simulcast_count:
            # 1 本になると simulcastEncodingAdapter がなくなる
            if simulcast_count > 1:
                assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
            assert encoder_implementation in s["encoderImplementation"]

            assert s["bytesSent"] > 500
            assert s["packetsSent"] > 5

            assert s["targetBitrate"] >= expect_target_bitrate(
                video_codec_type, s["frameWidth"], s["frameHeight"]
            ), f"{video_codec_type} {encoder_implementation} {video_bit_rate}"

            print(
                s["rid"],
                video_codec_type,
                encoder_implementation,
                video_bit_rate * 1000,
                video_width,
                video_height,
                s["targetBitrate"],
                s["frameWidth"],
                s["frameHeight"],
                s["bytesSent"],
                s["packetsSent"],
            )
        else:
            # 本来は 0 なのだが、simulcast_count が 1 の場合、
            # packetSent が 0 ではなく 1 や 2 になる場合がある
            # byteSent は 0
            assert s["bytesSent"] == 0
            assert s["packetsSent"] <= 2
            print(
                s["rid"],
                video_codec_type,
                s["bytesSent"],
                s["packetsSent"],
            )



import time

import pytest
from api import request_key_frame_api
from client import (
    SoraClient,
    SoraRole,
    codec_type_string_to_codec_type,
)

from sora_sdk import (
    SoraVideoCodecImplementation,
    SoraVideoCodecPreference,
)


@pytest.mark.parametrize(
    "video_codec_type",
    [
        "VP9",
        "VP8",
        "AV1",
    ],
)
def test_key_frame_request(settings, video_codec_type):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    encoder=SoraVideoCodecImplementation.INTERNAL,
                ),
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(3)

    assert sendonly.connection_id is not None

    # キーフレーム要求 API を 3 秒間隔で 3 回呼び出す
    api_count = 3
    for _ in range(api_count):
        response = request_key_frame_api(
            settings.api_url, sendonly.channel_id, sendonly.connection_id
        )
        assert response.status_code == 200
        time.sleep(3)

    # 統計を取得する
    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")

    # 3 回以上
    assert outbound_rtp_stats["keyFramesEncoded"] > api_count
    assert outbound_rtp_stats["pliCount"] >= api_count
    print("keyFramesEncoded:", outbound_rtp_stats["keyFramesEncoded"])
    print("pliCount:", outbound_rtp_stats["pliCount"])

    # PLI カウントの 50% 以上がキーフレームとしてエンコードされることを確認
    assert outbound_rtp_stats["keyFramesEncoded"] >= outbound_rtp_stats["pliCount"] * 0.7
    print(
        "keyFramesEncoded >= pliCount * 0.7:",
        outbound_rtp_stats["keyFramesEncoded"] >= outbound_rtp_stats["pliCount"] * 0.7,
    )



from client import SoraClient, SoraRole


def test_signaling_notify(settings):
    """
    2 接続での connection.created と connection.destroyed の通知を確認する
    """
    with SoraClient(
        settings,
        SoraRole.SENDRECV,
        audio=True,
        video=True,
    ) as c1:
        # c1 に自分の connection.created が通知される
        notify = c1.wait_notify(lambda notify: notify["event_type"] == "connection.created")
        assert notify["connection_id"] == c1.connection_id
        assert notify["channel_connections"] == 1

        with SoraClient(
            settings,
            SoraRole.SENDRECV,
            audio=True,
            video=True,
        ) as c2:
            # c2 に自分の connection.created が通知される
            notify = c2.wait_notify(lambda notify: notify["event_type"] == "connection.created")
            assert notify["connection_id"] == c2.connection_id
            assert notify["channel_connections"] == 2
            # data に c1 の connection_id が入ってる
            assert notify["data"][0]["connection_id"] == c1.connection_id

            # c1 に c2 の connection.created が通知される
            notify = c1.wait_notify(lambda notify: notify["event_type"] == "connection.created")
            assert notify["connection_id"] == c2.connection_id
            assert notify["channel_connections"] == 2

        # c1 に c2 の connection.destroyed が通知される
        notify = c1.wait_notify(lambda notify: notify["event_type"] == "connection.destroyed")
        assert notify["connection_id"] == c2.connection_id
        assert notify["channel_connections"] == 1



# buildbase.py はビルドスクリプトのテンプレートとなるファイル
#
# 自身のリポジトリにコピーして利用する。
#
# 元のファイルは以下のリポジトリにある:
# https://github.com/melpon/buildbase
#
# 更新する場合は以下のコマンドを利用する:
# curl -LO https://raw.githubusercontent.com/melpon/buildbase/master/buildbase.py
#
# ライセンス: Apache License 2.0
#
# Copyright 2024 melpon (Wandbox)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import filecmp
import glob
import hashlib
import logging
import multiprocessing
import os
import platform
import shlex
import shutil
import stat
import subprocess
import tarfile
import urllib.parse
import zipfile
from typing import Dict, List, NamedTuple, Optional

if platform.system() == "Windows":
    import winreg


class ChangeDirectory(object):
    def __init__(self, cwd):
        self._cwd = cwd

    def __enter__(self):
        self._old_cwd = os.getcwd()
        logging.debug(f"pushd {self._old_cwd} --> {self._cwd}")
        os.chdir(self._cwd)

    def __exit__(self, exctype, excvalue, trace):
        logging.debug(f"popd {self._old_cwd} <-- {self._cwd}")
        os.chdir(self._old_cwd)
        return False


def cd(cwd):
    return ChangeDirectory(cwd)


def cmd(args, **kwargs):
    logging.debug(f"+{args} {kwargs}")
    if "check" not in kwargs:
        kwargs["check"] = True
    if "resolve" in kwargs:
        resolve = kwargs["resolve"]
        del kwargs["resolve"]
    else:
        resolve = True
    if resolve:
        args = [shutil.which(args[0]), *args[1:]]
    return subprocess.run(args, **kwargs)


# 標準出力をキャプチャするコマンド実行。シェルの `cmd ...` や $(cmd ...) と同じ
def cmdcap(args, **kwargs):
    # 3.7 でしか使えない
    # kwargs['capture_output'] = True
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.PIPE
    kwargs["encoding"] = "utf-8"
    return cmd(args, **kwargs).stdout.strip()


# https://stackoverflow.com/a/2656405
def onerror(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.
    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.
    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    import stat

    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


def rm_rf(path: str):
    if not os.path.exists(path):
        logging.debug(f"rm -rf {path} => path not found")
        return
    if os.path.isfile(path) or os.path.islink(path):
        os.remove(path)
        logging.debug(f"rm -rf {path} => file removed")
    if os.path.isdir(path):
        shutil.rmtree(path, onerror=onerror)
        logging.debug(f"rm -rf {path} => directory removed")


def mkdir_p(path: str):
    if os.path.exists(path):
        logging.debug(f"mkdir -p {path} => already exists")
        return
    os.makedirs(path, exist_ok=True)
    logging.debug(f"mkdir -p {path} => directory created")


if platform.system() == "Windows":
    PATH_SEPARATOR = ";"
else:
    PATH_SEPARATOR = ":"


def add_path(path: str, is_after=False):
    logging.debug(f"add_path: {path}")
    if "PATH" not in os.environ:
        os.environ["PATH"] = path
        return

    if is_after:
        os.environ["PATH"] = os.environ["PATH"] + PATH_SEPARATOR + path
    else:
        os.environ["PATH"] = path + PATH_SEPARATOR + os.environ["PATH"]


def download(
    url: str,
    output_dir: Optional[str] = None,
    filename: Optional[str] = None,
    expected_sha256: Optional[str] = None,
) -> str:
    if filename is None:
        output_path = urllib.parse.urlparse(url).path.split("/")[-1]
    else:
        output_path = filename

    if output_dir is not None:
        output_path = os.path.join(output_dir, output_path)

    if os.path.exists(output_path):
        # ファイルが既に存在する場合でもハッシュチェックを行う
        if expected_sha256 is not None:
            try:
                verify_sha256(output_path, expected_sha256)
            except ValueError:
                # ハッシュ不一致の場合はファイルを削除して再ダウンロードを試みる
                logging.warning(f"Existing file has invalid hash, removing: {output_path}")
                os.remove(output_path)
                # 再ダウンロードを試みる
                return download(url, output_dir, filename, expected_sha256)
        else:
            return output_path

    try:
        logging.info(f"Downloading {url} to {output_path}")
        if shutil.which("curl") is not None:
            cmd(["curl", "-fLo", output_path, url])
        else:
            cmd(["wget", "-cO", output_path, url])

        # ダウンロード後にハッシュチェック
        if expected_sha256 is not None:
            try:
                verify_sha256(output_path, expected_sha256)
            except ValueError as e:
                logging.error(f"Hash verification failed. {e}")
                raise

    except Exception:
        # ゴミを残さないようにする
        if os.path.exists(output_path):
            logging.error(f"Removing incomplete/invalid file: {output_path}")
            os.remove(output_path)
        raise

    return output_path


def verify_sha256(file_path: str, expected_sha256: str):
    """ファイルの SHA256 ハッシュを検証する"""
    logging.info(f"Verifying SHA256 hash for {file_path}")
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        # メモリ効率のため、チャンクごとに読み込む
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    actual_sha256 = sha256_hash.hexdigest()
    if actual_sha256 != expected_sha256.lower():
        error_msg = (
            f"SHA256 hash mismatch for {file_path}:\n"
            f"  Expected: {expected_sha256.lower()}\n"
            f"  Actual:   {actual_sha256}"
        )
        logging.error(error_msg)
        logging.error("Aborting due to hash verification failure")
        raise ValueError(error_msg)
    else:
        logging.info(f"SHA256 hash verified successfully for {file_path}")


def read_version_file(path: str) -> Dict[str, str]:
    versions = {}

    lines = open(path, encoding="utf-8").readlines()
    for line in lines:
        line = line.strip()

        # コメント行
        if line[:1] == "#":
            continue

        # 空行
        if len(line) == 0:
            continue

        [a, b] = map(lambda x: x.strip(), line.split("=", 2))
        versions[a] = b.strip('"')

    return versions


# dir 以下にある全てのファイルパスを、dir2 からの相対パスで返す
def enum_all_files(dir, dir2):
    for root, _, files in os.walk(dir):
        for file in files:
            yield os.path.relpath(os.path.join(root, file), dir2)


def versioned(func):
    def wrapper(version, version_file, *args, **kwargs):
        if "ignore_version" in kwargs:
            if kwargs.get("ignore_version"):
                rm_rf(version_file)
            del kwargs["ignore_version"]

        if os.path.exists(version_file):
            ver = open(version_file, encoding="utf-8").read()
            if ver.strip() == version.strip():
                return

        r = func(version=version, *args, **kwargs)

        with open(version_file, "w", encoding="utf-8") as f:
            f.write(version)

        return r

    return wrapper


# アーカイブが単一のディレクトリに全て格納されているかどうかを調べる。
#
# 単一のディレクトリに格納されている場合はそのディレクトリ名を返す。
# そうでない場合は None を返す。
def _is_single_dir(infos, get_name, is_dir) -> Optional[str]:
    # tarfile: ['path', 'path/to', 'path/to/file.txt']
    # zipfile: ['path/', 'path/to/', 'path/to/file.txt']
    # どちらも / 区切りだが、ディレクトリの場合、後ろに / が付くかどうかが違う
    dirname = None
    for info in infos:
        name = get_name(info)
        n = name.rstrip("/").find("/")
        if n == -1:
            # ルートディレクトリにファイルが存在している
            if not is_dir(info):
                return None
            dir = name.rstrip("/")
        else:
            dir = name[0:n]
        # ルートディレクトリに２個以上のディレクトリが存在している
        if dirname is not None and dirname != dir:
            return None
        dirname = dir

    return dirname


def is_single_dir_tar(tar: tarfile.TarFile) -> Optional[str]:
    return _is_single_dir(tar.getmembers(), lambda t: t.name, lambda t: t.isdir())


def is_single_dir_zip(zip: zipfile.ZipFile) -> Optional[str]:
    return _is_single_dir(zip.infolist(), lambda z: z.filename, lambda z: z.is_dir())


# 解凍した上でファイル属性を付与する
def _extractzip(z: zipfile.ZipFile, path: str):
    z.extractall(path)
    if platform.system() == "Windows":
        return
    for info in z.infolist():
        if info.is_dir():
            continue
        filepath = os.path.join(path, info.filename)
        mod = info.external_attr >> 16
        if (mod & 0o120000) == 0o120000:
            # シンボリックリンク
            with open(filepath, "r", encoding="utf-8") as f:
                src = f.read()
            os.remove(filepath)
            with cd(os.path.dirname(filepath)):
                if os.path.exists(src):
                    os.symlink(src, filepath)
        if os.path.exists(filepath):
            # 普通のファイル
            os.chmod(filepath, mod & 0o777)


# zip または tar.gz ファイルを展開する。
#
# 展開先のディレクトリは {output_dir}/{output_dirname} となり、
# 展開先のディレクトリが既に存在していた場合は削除される。
#
# もしアーカイブの内容が単一のディレクトリであった場合、
# そのディレクトリは無いものとして展開される。
#
# つまりアーカイブ libsora-1.23.tar.gz の内容が
# ['libsora-1.23', 'libsora-1.23/file1', 'libsora-1.23/file2']
# であった場合、extract('libsora-1.23.tar.gz', 'out', 'libsora') のようにすると
# - out/libsora/file1
# - out/libsora/file2
# が出力される。
#
# また、アーカイブ libsora-1.23.tar.gz の内容が
# ['libsora-1.23', 'libsora-1.23/file1', 'libsora-1.23/file2', 'LICENSE']
# であった場合、extract('libsora-1.23.tar.gz', 'out', 'libsora') のようにすると
# - out/libsora/libsora-1.23/file1
# - out/libsora/libsora-1.23/file2
# - out/libsora/LICENSE
# が出力される。
def extract(file: str, output_dir: str, output_dirname: str, filetype: Optional[str] = None):
    path = os.path.join(output_dir, output_dirname)
    logging.info(f"Extract {file} to {path}")
    if (
        filetype == "gzip"
        or file.endswith(".tar.gz")
        or filetype == "lzma"
        or file.endswith(".tar.xz")
    ):
        rm_rf(path)
        with tarfile.open(file) as t:
            dir = is_single_dir_tar(t)
            if dir is None:
                os.makedirs(path, exist_ok=True)
                t.extractall(path)
            else:
                logging.info(f"Directory {dir} is stripped")
                path2 = os.path.join(output_dir, dir)
                rm_rf(path2)
                t.extractall(output_dir)
                if path != path2:
                    logging.debug(f"mv {path2} {path}")
                    os.replace(path2, path)
    elif filetype == "zip" or file.endswith(".zip"):
        rm_rf(path)
        with zipfile.ZipFile(file) as z:
            dir = is_single_dir_zip(z)
            if dir is None:
                os.makedirs(path, exist_ok=True)
                # z.extractall(path)
                _extractzip(z, path)
            else:
                logging.info(f"Directory {dir} is stripped")
                path2 = os.path.join(output_dir, dir)
                rm_rf(path2)
                # z.extractall(output_dir)
                _extractzip(z, output_dir)
                if path != path2:
                    logging.debug(f"mv {path2} {path}")
                    os.replace(path2, path)
    else:
        raise Exception("file should end with .tar.gz or .zip")


def clone_and_checkout(url, version, dir, fetch, fetch_force):
    if fetch_force:
        rm_rf(dir)

    if not os.path.exists(os.path.join(dir, ".git")):
        cmd(["git", "clone", url, dir])
        fetch = True

    if fetch:
        with cd(dir):
            cmd(["git", "fetch"])
            cmd(["git", "reset", "--hard"])
            cmd(["git", "clean", "-df"])
            cmd(["git", "checkout", "-f", version])


def git_clone_shallow(url, hash, dir, submodule=False):
    rm_rf(dir)
    mkdir_p(dir)
    with cd(dir):
        cmd(["git", "init"])
        cmd(["git", "remote", "add", "origin", url])
        cmd(["git", "fetch", "--depth=1", "origin", hash])
        cmd(["git", "reset", "--hard", "FETCH_HEAD"])
        if submodule:
            cmd(
                [
                    "git",
                    "submodule",
                    "update",
                    "--init",
                    "--recursive",
                    "--recommend-shallow",
                    "--depth",
                    "1",
                ]
            )


def apply_patch(patch, dir, depth):
    patch = os.path.abspath(patch)
    with cd(dir):
        logging.info(f"patch -p{depth} < {patch}")
        if platform.system() == "Windows":
            cmd(
                [
                    "git",
                    "apply",
                    f"-p{depth}",
                    "--ignore-space-change",
                    "--ignore-whitespace",
                    "--whitespace=nowarn",
                    "--reject",
                    patch,
                ]
            )
        else:
            with open(patch, encoding="utf-8") as stdin:
                cmd(["patch", f"-p{depth}"], stdin=stdin)


def apply_patch_text(patch_text, dir, depth):
    with cd(dir):
        logging.info(f"echo '{patch_text[:100]}...' | patch -p{depth} -")
        directory = cmdcap(["git", "rev-parse", "--show-prefix"])
        if platform.system() == "Windows":
            cmd(
                [
                    "git",
                    "apply",
                    f"-p{depth}",
                    "--ignore-space-change",
                    "--ignore-whitespace",
                    "--whitespace=nowarn",
                    "--reject",
                    f"--directory={directory}",
                    "-",
                ],
                input=patch_text,
                text=True,
                encoding="utf-8",
            )
        else:
            cmd(["patch", f"-p{depth}"], input=patch_text, text=True, encoding="utf-8")


def copyfile_if_different(src, dst):
    if os.path.exists(dst) and filecmp.cmp(src, dst, shallow=False):
        return
    shutil.copyfile(src, dst)


# NOTE(enm10k): shutil.copytree に Python 3.8 で追加された dirs_exist_ok=True を指定して使いたかったが、
# GitHub Actions の Windows のランナー (widnwos-2019) にインストールされている Python のバージョンが古くて利用できなかった
# actions/setup-python で Python 3.8 を設定してビルドしたところ、 Lyra のビルドがエラーになったためこの関数を自作した
# Windows のランナーを更新した場合は、この関数は不要になる可能性が高い
def copytree(src_dir, dst_dir):
    for file_path in glob.glob(src_dir + "/**", recursive=True):
        dest_path = os.path.join(dst_dir, os.path.relpath(file_path, src_dir))

        if os.path.isdir(file_path):
            os.makedirs(dest_path, exist_ok=True)
        else:
            shutil.copy2(file_path, dest_path)


def git_get_url_and_revision(dir):
    with cd(dir):
        rev = cmdcap(["git", "rev-parse", "HEAD"])
        url = cmdcap(["git", "remote", "get-url", "origin"])
        return url, rev


def replace_vcproj_static_runtime(project_file: str):
    # なぜか MSVC_STATIC_RUNTIME が効かずに DLL ランタイムを使ってしまうので
    # 生成されたプロジェクトに対して静的ランタイムを使うように変更する
    s = open(project_file, "r", encoding="utf-8").read()
    s = s.replace("MultiThreadedDLL", "MultiThreaded")
    s = s.replace("MultiThreadedDebugDLL", "MultiThreadedDebug")
    open(project_file, "w", encoding="utf-8").write(s)


@versioned
def install_webrtc(version, source_dir, install_dir, platform: str):
    win = platform.startswith("windows_")
    filename = f"webrtc.{platform}.{'zip' if win else 'tar.gz'}"
    rm_rf(os.path.join(source_dir, filename))
    archive = download(
        f"https://github.com/shiguredo-webrtc-build/webrtc-build/releases/download/{version}/{filename}",
        output_dir=source_dir,
    )
    rm_rf(os.path.join(install_dir, "webrtc"))
    extract(archive, output_dir=install_dir, output_dirname="webrtc")


def build_webrtc(platform, local_webrtc_build_dir, local_webrtc_build_args, debug):
    with cd(local_webrtc_build_dir):
        args = ["--webrtc-nobuild-ios-framework", "--webrtc-nobuild-android-aar"]
        if debug:
            args += ["--debug"]

        args += local_webrtc_build_args

        cmd(["python3", "run.py", "build", platform, *args])

        # インクルードディレクトリを増やしたくないので、
        # __config_site を libc++ のディレクトリにコピーしておく
        webrtc_source_dir = os.path.join(local_webrtc_build_dir, "_source", platform, "webrtc")
        src_config = os.path.join(
            webrtc_source_dir, "src", "buildtools", "third_party", "libc++", "__config_site"
        )
        dst_config = os.path.join(
            webrtc_source_dir, "src", "third_party", "libc++", "src", "include", "__config_site"
        )
        copyfile_if_different(src_config, dst_config)

        # __assertion_handler をコピーする
        src_assertion = os.path.join(
            webrtc_source_dir,
            "src",
            "buildtools",
            "third_party",
            "libc++",
            "__assertion_handler",
        )
        dst_assertion = os.path.join(
            webrtc_source_dir,
            "src",
            "third_party",
            "libc++",
            "src",
            "include",
            "__assertion_handler",
        )
        copyfile_if_different(src_assertion, dst_assertion)


class WebrtcInfo(NamedTuple):
    version_file: str
    deps_file: str
    webrtc_include_dir: str
    webrtc_source_dir: Optional[str]
    webrtc_library_dir: str
    clang_dir: str
    libcxx_dir: str
    libcxxabi_dir: str


def get_webrtc_info(
    platform: str, local_webrtc_build_dir: Optional[str], install_dir: str, debug: bool
) -> WebrtcInfo:
    webrtc_install_dir = os.path.join(install_dir, "webrtc")

    if local_webrtc_build_dir is None:
        return WebrtcInfo(
            version_file=os.path.join(webrtc_install_dir, "VERSIONS"),
            deps_file=os.path.join(webrtc_install_dir, "DEPS"),
            webrtc_include_dir=os.path.join(webrtc_install_dir, "include"),
            webrtc_source_dir=None,
            webrtc_library_dir=os.path.join(webrtc_install_dir, "lib"),
            clang_dir=os.path.join(install_dir, "llvm", "clang"),
            libcxx_dir=os.path.join(install_dir, "llvm", "libcxx"),
            libcxxabi_dir=os.path.join(
                webrtc_install_dir, "include", "third_party", "libc++abi", "src"
            ),
        )
    else:
        webrtc_build_source_dir = os.path.join(
            local_webrtc_build_dir, "_source", platform, "webrtc"
        )
        configuration = "debug" if debug else "release"
        webrtc_build_build_dir = os.path.join(
            local_webrtc_build_dir, "_build", platform, configuration, "webrtc"
        )

        return WebrtcInfo(
            version_file=os.path.join(local_webrtc_build_dir, "VERSION"),
            deps_file=os.path.join(local_webrtc_build_dir, "DEPS"),
            webrtc_include_dir=os.path.join(webrtc_build_source_dir, "src"),
            webrtc_source_dir=os.path.join(webrtc_build_source_dir, "src"),
            webrtc_library_dir=webrtc_build_build_dir,
            clang_dir=os.path.join(
                webrtc_build_source_dir, "src", "third_party", "llvm-build", "Release+Asserts"
            ),
            libcxx_dir=os.path.join(webrtc_build_source_dir, "src", "third_party", "libc++", "src"),
            libcxxabi_dir=os.path.join(
                webrtc_build_source_dir, "src", "third_party", "libc++abi", "src"
            ),
        )


@versioned
def install_boost(
    version,
    source_dir,
    install_dir,
    sora_version,
    platform: str,
    expected_sha256: Optional[str] = None,
):
    win = platform.startswith("windows_")
    filename = (
        f"boost-{version}_sora-cpp-sdk-{sora_version}_{platform}.{'zip' if win else 'tar.gz'}"
    )
    rm_rf(os.path.join(source_dir, filename))
    archive = download(
        f"https://github.com/shiguredo/sora-cpp-sdk/releases/download/{sora_version}/{filename}",
        output_dir=source_dir,
        expected_sha256=expected_sha256,
    )
    rm_rf(os.path.join(install_dir, "boost"))
    extract(archive, output_dir=install_dir, output_dirname="boost")


# 以下の問題を解決するためのパッチ
#
# No support for msvc-toolset 14.4x (VS 2022, 17.10.x): https://github.com/boostorg/boost/issues/914
BOOST_PATCH_SUPPORT_14_4 = r"""
diff --git a/tools/build/src/engine/config_toolset.bat b/tools/build/src/engine/config_toolset.bat
index 4ba577cac..3e3f6a3a1 100644
--- a/tools/build/src/engine/config_toolset.bat
+++ b/tools/build/src/engine/config_toolset.bat
@@ -157,7 +157,7 @@ pushd %CD%
 if "_%VSINSTALLDIR%_" == "__" call :Call_If_Exists "%B2_TOOLSET_ROOT%Auxiliary\Build\vcvarsall.bat" %B2_BUILD_ARGS%
 popd
 @REM set "B2_CXX="%CXX%" /nologo /MP /MT /TP /Feb2 /wd4996 /O2 /GL /EHsc"
-set "B2_CXX="%CXX%" /nologo -TP /wd4996 /wd4675 /EHs /GR /Zc:throwingNew /O2 /Ob2 /W3 /MD /Zc:forScope /Zc:wchar_t /Zc:inline /Gw /favor:blend /Feb2"
+set "B2_CXX="%CXX%" /nologo -TP /wd4996 /wd4675 /EHs /GR /Zc:throwingNew /O2 /Ob2 /W3 /MT /Zc:forScope /Zc:wchar_t /Zc:inline /Gw /favor:blend /Feb2"
 set "B2_CXX_LINK=/link kernel32.lib advapi32.lib user32.lib"
 set "_known_=1"
 goto :Embed_Minafest_Via_Link
diff --git a/tools/build/src/tools/msvc.jam b/tools/build/src/tools/msvc.jam
index 54a6ced32..4bb3810b3 100644
--- a/tools/build/src/tools/msvc.jam
+++ b/tools/build/src/tools/msvc.jam
@@ -1137,7 +1137,15 @@ local rule generate-setup-cmd ( version : command : parent : options * : cpu : g
         }
         else
         {
-            if [ MATCH "(14.3)" : $(version) ]
+            if [ MATCH "(14.4)" : $(version) ]
+            {
+                if $(.debug-configuration)
+                {
+                    ECHO "notice: [generate-setup-cmd] $(version) is 14.4" ;
+                }
+                parent = [ path.native [ path.join  $(parent) "..\\..\\..\\..\\..\\Auxiliary\\Build" ] ] ;
+            }
+            else if [ MATCH "(14.3)" : $(version) ]
             {
                 if $(.debug-configuration)
                 {
@@ -1316,7 +1324,11 @@ local rule configure-really ( version ? : options * )
             # version from the path.
             # FIXME: We currently detect both Microsoft Visual Studio 9.0 and
             # 9.0express as 9.0 here.
-            if [ MATCH "(MSVC\\\\14.3)" : $(command) ]
+            if [ MATCH "(MSVC\\\\14.4)" : $(command) ]
+            {
+                version = 14.4 ;
+            }
+            else if [ MATCH "(MSVC\\\\14.3)" : $(command) ]
             {
                 version = 14.3 ;
             }
@@ -1745,13 +1757,17 @@ local rule default-path ( version )
         # And fortunately, forward slashes do also work in native Windows.
         local vswhere = "$(root)/Microsoft Visual Studio/Installer/vswhere.exe" ;
         # The check for $(root) is to avoid a segmentation fault if not found.
-        if $(version) in 14.1 14.2 14.3 default && $(root) && [ path.exists $(vswhere) ]
+        if $(version) in 14.1 14.2 14.3 14.4 default && $(root) && [ path.exists $(vswhere) ]
         {
             local req = "-requires Microsoft.VisualStudio.Component.VC.Tools.x86.x64" ;
             local prop = "-property installationPath" ;
             local limit ;
 
-            if $(version) = 14.3
+            if $(version) = 14.4
+            {
+                limit = "-version \"[17.0,18.0)\" -prerelease" ;
+            }
+            else if $(version) = 14.3
             {
                 limit = "-version \"[17.0,18.0)\" -prerelease" ;
             }
@@ -2174,7 +2190,7 @@ for local arch in [ MATCH "^\\.cpus-on-(.*)" : [ VARNAMES $(__name__) ] ]
                      armv7 armv7s ;
 
 # Known toolset versions, in order of preference.
-.known-versions = 14.3 14.2 14.1 14.0 12.0 11.0 10.0 10.0express 9.0 9.0express 8.0 8.0express 7.1
+.known-versions = 14.4 14.3 14.2 14.1 14.0 12.0 11.0 10.0 10.0express 9.0 9.0express 8.0 8.0express 7.1
     7.1toolkit 7.0 6.0 ;
 
 # Version aliases.
@@ -2226,6 +2242,11 @@ for local arch in [ MATCH "^\\.cpus-on-(.*)" : [ VARNAMES $(__name__) ] ]
     "Microsoft Visual Studio/2022/*/VC/Tools/MSVC/*/bin/Host*/*"
     ;
 .version-14.3-env = VS170COMNTOOLS ProgramFiles ProgramFiles(x86) ;
+.version-14.4-path =
+    "../../VC/Tools/MSVC/*/bin/Host*/*"
+    "Microsoft Visual Studio/2022/*/VC/Tools/MSVC/*/bin/Host*/*"
+    ;
+.version-14.4-env = VS170COMNTOOLS ProgramFiles ProgramFiles(x86) ;
 
 # Auto-detect all the available msvc installations on the system.
 auto-detect-toolset-versions ;
"""


@versioned
def build_and_install_boost(
    version: str,
    source_dir,
    build_dir,
    install_dir,
    debug: bool,
    cxx: str,
    cflags: List[str],
    cxxflags: List[str],
    linkflags: List[str],
    toolset,
    visibility,
    target_os,
    architecture,
    android_ndk,
    native_api_level,
    address_model="64",
    runtime_link=None,
    android_build_platform="linux-x86_64",
    expected_sha256: Optional[str] = None,
):
    version_underscore = version.replace(".", "_")

    archive = download(
        # 公式サイトに負荷をかけないための時雨堂によるミラー
        f"https://oss-mirrors.shiguredo.jp/boost_{version_underscore}.tar.gz",
        # Boost 公式のミラー
        # f"https://archives.boost.io/release/{version}/source/boost_{version_underscore}.tar.gz",
        source_dir,
        expected_sha256=expected_sha256,
    )
    extract(archive, output_dir=build_dir, output_dirname="boost")
    with cd(os.path.join(build_dir, "boost")):
        if target_os == "windows":
            bootstrap = ".\\bootstrap.bat"
            b2 = "b2"
        elif target_os == "android" and android_build_platform == "windows-x86_64":
            # Android を Windows でビルドする場合
            bootstrap = ".\\bootstrap.bat"
            b2 = "b2"
        else:
            bootstrap = "./bootstrap.sh"
            b2 = "./b2"

        if runtime_link is None:
            runtime_link = "static" if target_os == "windows" else "shared"

        # Windows かつ Boost 1.85.0 の場合はパッチを当てる
        if target_os == "windows" and version == "1.85.0":
            apply_patch_text(BOOST_PATCH_SUPPORT_14_4, os.path.join(build_dir, "boost"), 1)

        cmd([bootstrap])

        if target_os == "iphone":
            IOS_BUILD_TARGETS = [("arm64", "iphoneos")]
            for arch, sdk in IOS_BUILD_TARGETS:
                # xcode の clang++ を利用する
                # ただし cxx が指定されてた場合はそちらを優先する
                clangpp = cmdcap(["xcodebuild", "-find", "clang++"]) if len(cxx) == 0 else cxx
                sysroot = cmdcap(["xcrun", "--sdk", sdk, "--show-sdk-path"])
                boost_arch = "x86" if arch == "x86_64" else "arm"
                with open("project-config.jam", "w", encoding="utf-8") as f:
                    f.write(
                        f"using clang \
                        : iphone \
                        : {clangpp} -arch {arch} -isysroot {sysroot} \
                          -fembed-bitcode \
                          -mios-version-min=10.0 \
                          -fvisibility=hidden \
                        : <striper> <root>{sysroot} \
                        ; \
                        "
                    )
                cmd(
                    [
                        b2,
                        "install",
                        "-d+0",
                        f"--build-dir={os.path.join(build_dir, 'boost', f'build-{arch}-{sdk}')}",
                        f"--prefix={os.path.join(build_dir, 'boost', f'install-{arch}-{sdk}')}",
                        "--with-json",
                        "--with-filesystem",
                        "--layout=system",
                        "--ignore-site-config",
                        f"variant={'debug' if debug else 'release'}",
                        f"cflags={' '.join(cflags)}",
                        f"cxxflags={' '.join(cxxflags)}",
                        f"linkflags={' '.join(linkflags)}",
                        f"toolset={toolset}",
                        f"visibility={visibility}",
                        f"target-os={target_os}",
                        f"address-model={address_model}",
                        "link=static",
                        f"runtime-link={runtime_link}",
                        "threading=multi",
                        f"architecture={boost_arch}",
                    ]
                )
            arch, sdk = IOS_BUILD_TARGETS[0]
            installed_path = os.path.join(build_dir, "boost", f"install-{arch}-{sdk}")
            rm_rf(os.path.join(install_dir, "boost"))
            cmd(["cp", "-r", installed_path, os.path.join(install_dir, "boost")])

            for lib in enum_all_files(
                os.path.join(installed_path, "lib"), os.path.join(installed_path, "lib")
            ):
                if not lib.endswith(".a"):
                    continue
                files = [
                    os.path.join(build_dir, "boost", f"install-{arch}-{sdk}", "lib", lib)
                    for arch, sdk in IOS_BUILD_TARGETS
                ]
                if len(files) == 1:
                    shutil.copyfile(files[0], os.path.join(install_dir, "boost", "lib", lib))
                else:
                    cmd(
                        [
                            "lipo",
                            "-create",
                            "-output",
                            os.path.join(install_dir, "boost", "lib", lib),
                        ]
                        + files
                    )
        elif target_os == "android":
            # Android の場合、android-ndk を使ってビルドする
            # ただし cxx が指定されてた場合はそちらを優先する
            with open("project-config.jam", "w", encoding="utf-8") as f:
                bin = os.path.join(
                    android_ndk, "toolchains", "llvm", "prebuilt", android_build_platform, "bin"
                )
                sysroot = os.path.join(
                    android_ndk, "toolchains", "llvm", "prebuilt", android_build_platform, "sysroot"
                )

                def escape(s):
                    return s.replace("\\", "/").replace(":", "\\:")

                f.write(
                    f"using clang \
                    : android \
                    : {escape(cxx if len(cxx) != 0 else os.path.join(bin, 'clang++'))} \
                      --target=aarch64-none-linux-android{native_api_level} \
                      --sysroot={escape(sysroot)} \
                    : <archiver>{escape(os.path.join(bin, 'llvm-ar'))} \
                      <ranlib>{escape(os.path.join(bin, 'llvm-ranlib'))} \
                    ; \
                    "
                )
            cmd(
                [
                    b2,
                    "install",
                    "-d+0",
                    f"--prefix={os.path.join(install_dir, 'boost')}",
                    "--with-json",
                    "--with-filesystem",
                    "--layout=system",
                    "--ignore-site-config",
                    f"variant={'debug' if debug else 'release'}",
                    f"compileflags=--sysroot={sysroot}",
                    f"cflags={' '.join(cflags)}",
                    f"cxxflags={' '.join(cxxflags)}",
                    f"linkflags={' '.join(linkflags)}",
                    f"toolset={toolset}",
                    f"visibility={visibility}",
                    f"target-os={target_os}",
                    f"address-model={address_model}",
                    "link=static",
                    f"runtime-link={runtime_link}",
                    "threading=multi",
                    "architecture=arm",
                ]
            )
        else:
            if len(cxx) != 0:
                with open("project-config.jam", "w", encoding="utf-8") as f:
                    f.write(f"using {toolset} : : {cxx} : ;")
            cmd(
                [
                    b2,
                    "install",
                    "-d+0",
                    f"--prefix={os.path.join(install_dir, 'boost')}",
                    "--with-json",
                    "--with-filesystem",
                    "--layout=system",
                    "--ignore-site-config",
                    f"variant={'debug' if debug else 'release'}",
                    f"cflags={' '.join(cflags)}",
                    f"cxxflags={' '.join(cxxflags)}",
                    f"linkflags={' '.join(linkflags)}",
                    f"toolset={toolset}",
                    f"visibility={visibility}",
                    f"target-os={target_os}",
                    f"address-model={address_model}",
                    "link=static",
                    f"runtime-link={runtime_link}",
                    "threading=multi",
                    f"architecture={architecture}",
                ]
            )


@versioned
def install_sora(version, source_dir, install_dir, platform: str):
    win = platform.startswith("windows_")
    filename = f"sora-cpp-sdk-{version}_{platform}.{'zip' if win else 'tar.gz'}"
    rm_rf(os.path.join(source_dir, filename))
    archive = download(
        f"https://github.com/shiguredo/sora-cpp-sdk/releases/download/{version}/{filename}",
        output_dir=source_dir,
    )
    rm_rf(os.path.join(install_dir, "sora"))
    extract(archive, output_dir=install_dir, output_dirname="sora")


def install_sora_and_deps(
    sora_version: str, boost_version: str, platform: str, source_dir: str, install_dir: str
):
    # Boost
    install_boost_args = {
        "version": boost_version,
        "version_file": os.path.join(install_dir, "boost.version"),
        "source_dir": source_dir,
        "install_dir": install_dir,
        "sora_version": sora_version,
        "platform": platform,
    }
    install_boost(**install_boost_args)

    # Sora C++ SDK
    install_sora_args = {
        "version": sora_version,
        "version_file": os.path.join(install_dir, "sora.version"),
        "source_dir": source_dir,
        "install_dir": install_dir,
        "platform": platform,
    }
    install_sora(**install_sora_args)


def build_sora(
    platform: str,
    local_sora_cpp_sdk_dir: str,
    local_sora_cpp_sdk_args: List[str],
    debug: bool,
    local_webrtc_build_dir: Optional[str],
):
    if debug and "--debug" not in local_sora_cpp_sdk_args:
        local_sora_cpp_sdk_args = ["--debug", *local_sora_cpp_sdk_args]
    if local_webrtc_build_dir is not None:
        local_sora_cpp_sdk_args = [
            "--local-webrtc-build-dir",
            local_webrtc_build_dir,
            *local_sora_cpp_sdk_args,
        ]

    with cd(local_sora_cpp_sdk_dir):
        cmd(["python3", "run.py", "build", platform, *local_sora_cpp_sdk_args])


class SoraInfo(NamedTuple):
    sora_install_dir: str
    boost_install_dir: str


def get_sora_info(
    platform: str, local_sora_cpp_sdk_dir: Optional[str], install_dir: str, debug: bool
) -> SoraInfo:
    if local_sora_cpp_sdk_dir is not None:
        configuration = "debug" if debug else "release"
        install_dir = os.path.join(local_sora_cpp_sdk_dir, "_install", platform, configuration)

    return SoraInfo(
        sora_install_dir=os.path.join(install_dir, "sora"),
        boost_install_dir=os.path.join(install_dir, "boost"),
    )


@versioned
def install_rootfs(version, install_dir, conf, arch="arm64"):
    rootfs_dir = os.path.join(install_dir, "rootfs")
    rm_rf(rootfs_dir)
    cmd(["multistrap", "--no-auth", "-a", arch, "-d", rootfs_dir, "-f", conf])
    # 絶対パスのシンボリックリンクを相対パスに置き換えていく
    for dir, _, filenames in os.walk(rootfs_dir):
        for filename in filenames:
            linkpath = os.path.join(dir, filename)
            # symlink かどうか
            if not os.path.islink(linkpath):
                continue
            target = os.readlink(linkpath)
            # 絶対パスかどうか
            if not os.path.isabs(target):
                continue
            # rootfs_dir を先頭に付けることで、
            # rootfs の外から見て正しい絶対パスにする
            targetpath = rootfs_dir + target
            # 参照先の絶対パスが存在するかどうか
            if not os.path.exists(targetpath):
                continue
            # 相対パスに置き換える
            relpath = os.path.relpath(targetpath, dir)
            logging.debug(f"{linkpath[len(rootfs_dir) :]} targets {target} to {relpath}")
            os.remove(linkpath)
            os.symlink(relpath, linkpath)

    # なぜかシンボリックリンクが登録されていないので作っておく
    link = os.path.join(rootfs_dir, "usr", "lib", "aarch64-linux-gnu", "tegra", "libnvbuf_fdmap.so")
    file = os.path.join(
        rootfs_dir, "usr", "lib", "aarch64-linux-gnu", "tegra", "libnvbuf_fdmap.so.1.0.0"
    )
    if os.path.exists(file) and not os.path.exists(link):
        os.symlink(os.path.basename(file), link)

    # JetPack 6 から tegra → nvidia になった
    link = os.path.join(
        rootfs_dir, "usr", "lib", "aarch64-linux-gnu", "nvidia", "libnvbuf_fdmap.so"
    )
    file = os.path.join(
        rootfs_dir, "usr", "lib", "aarch64-linux-gnu", "nvidia", "libnvbuf_fdmap.so.1.0.0"
    )
    if os.path.exists(file) and not os.path.exists(link):
        os.symlink(os.path.basename(file), link)


@versioned
def install_android_ndk(version, install_dir, source_dir, platform="linux"):
    if platform not in ("windows", "darwin", "linux"):
        raise Exception(f"Not supported platform: {platform}")

    if platform == "windows":
        url = f"https://dl.google.com/android/repository/android-ndk-{version}-{platform}.zip"
    elif platform == "darwin":
        url = f"https://dl.google.com/android/repository/android-ndk-{version}-{platform}.dmg"
    else:
        url = f"https://dl.google.com/android/repository/android-ndk-{version}-{platform}.zip"
    archive = download(url, source_dir)
    rm_rf(os.path.join(install_dir, "android-ndk"))
    if platform == "darwin":
        cap = cmdcap(["hdiutil", "attach", archive])
        # 以下のような結果が得られるはずなので、ここから /Volumes/Android NDK r26 のところだけ取り出す
        # /dev/disk4              GUID_partition_scheme
        # /dev/disk4s1            EFI
        # /dev/disk4s2            Apple_HFS                       /Volumes/Android NDK r26
        volume = cap.split("\n")[-1].split("\t")[-1]
        # AndroidNDK10792818.app みたいな感じの app があるはず
        app = glob.glob("AndroidNDK*.app", root_dir=volume)[0]
        # NDK ディレクトリをコピー
        cmd(
            [
                "cp",
                "-r",
                os.path.join(volume, app, "Contents", "NDK"),
                os.path.join(install_dir, "android-ndk"),
            ]
        )
        cmdcap(["hdiutil", "detach", volume])
    else:
        extract(archive, output_dir=install_dir, output_dirname="android-ndk")


@versioned
def install_android_sdk_cmdline_tools(version, install_dir, source_dir):
    archive = download(
        f"https://dl.google.com/android/repository/commandlinetools-linux-{version}_latest.zip",
        source_dir,
    )
    tools_dir = os.path.join(install_dir, "android-sdk-cmdline-tools")
    rm_rf(tools_dir)
    extract(archive, output_dir=tools_dir, output_dirname="cmdline-tools")
    sdkmanager = os.path.join(tools_dir, "cmdline-tools", "bin", "sdkmanager")
    # ライセンスを許諾する
    cmd(["/bin/bash", "-c", f"yes | {sdkmanager} --sdk_root={tools_dir} --licenses"])


@versioned
def install_android_sdk_platform_tools(version, install_dir, source_dir, platform):
    if version not in ("latest",):
        raise Exception(f"Supports only 'latest' version, but got: {version}")
    if platform not in ("windows", "darwin", "linux"):
        raise Exception(f"Not supported platform: {platform}")
    archive = download(
        f"https://dl.google.com/android/repository/platform-tools-{version}-{platform}.zip",
        source_dir,
    )
    tools_dir = os.path.join(install_dir, "android-sdk-platform-tools")
    rm_rf(tools_dir)
    extract(archive, output_dir=tools_dir, output_dirname="platform-tools")


@versioned
def install_llvm(
    version,
    install_dir,
    tools_url,
    tools_commit,
    libcxx_url,
    libcxx_commit,
    buildtools_url,
    buildtools_commit,
):
    llvm_dir = os.path.join(install_dir, "llvm")
    rm_rf(llvm_dir)
    mkdir_p(llvm_dir)
    with cd(llvm_dir):
        # tools の update.py を叩いて特定バージョンの clang バイナリを拾う
        git_clone_shallow(tools_url, tools_commit, "tools")
        with cd("tools"):
            cmd(
                [
                    "python3",
                    os.path.join("clang", "scripts", "update.py"),
                    "--output-dir",
                    os.path.join(llvm_dir, "clang"),
                ]
            )

        # 特定バージョンの libcxx を利用する
        git_clone_shallow(libcxx_url, libcxx_commit, "libcxx")

        # __config_site のために特定バージョンの buildtools を取得する
        git_clone_shallow(buildtools_url, buildtools_commit, "buildtools")
        with cd("buildtools"):
            cmd(["git", "reset", "--hard", buildtools_commit])
        shutil.copyfile(
            os.path.join(llvm_dir, "buildtools", "third_party", "libc++", "__config_site"),
            os.path.join(llvm_dir, "libcxx", "include", "__config_site"),
        )

        # __assertion_handler をコピーする
        # 背景: https://source.chromium.org/chromium/_/chromium/external/github.com/llvm/llvm-project/libcxx.git/+/1e5bda0d1ce8e346955aa4a85eaab258785f11f7
        shutil.copyfile(
            # NOTE(enm10k): 最初は default_assertion_handler.in をコピーしていたが、 buildtools 以下に
            # default_assertion_handler.in から生成されたと思われる __assertion_handler が存在するため、それをコピーする
            # os.path.join(llvm_dir, "libcxx", "vendor", "llvm", "default_assertion_handler.in"),
            os.path.join(llvm_dir, "buildtools", "third_party", "libc++", "__assertion_handler"),
            os.path.join(llvm_dir, "libcxx", "include", "__assertion_handler"),
        )


def cmake_path(path: str) -> str:
    return path.replace("\\", "/")


@versioned
def install_cmake(version, source_dir, install_dir, platform: str, ext):
    url = f"https://github.com/Kitware/CMake/releases/download/v{version}/cmake-{version}-{platform}.{ext}"
    path = download(url, source_dir)
    extract(path, install_dir, "cmake")
    # Android で自前の CMake を利用する場合、ninja へのパスが見つけられない問題があるので、同じディレクトリに symlink を貼る
    # https://issuetracker.google.com/issues/206099937
    if platform.startswith("linux"):
        with cd(os.path.join(install_dir, "cmake", "bin")):
            cmd(["ln", "-s", "/usr/bin/ninja", "ninja"])


@versioned
def install_sdl2(
    version, source_dir, build_dir, install_dir, debug: bool, platform: str, cmake_args: List[str]
):
    url = (
        f"https://github.com/libsdl-org/SDL/releases/download/release-{version}/SDL2-{version}.zip"
    )
    # url = f"http://www.libsdl.org/release/SDL2-{version}.zip"
    path = download(url, source_dir)
    sdl2_source_dir = os.path.join(source_dir, "sdl2")
    sdl2_build_dir = os.path.join(build_dir, "sdl2")
    sdl2_install_dir = os.path.join(install_dir, "sdl2")
    rm_rf(sdl2_source_dir)
    rm_rf(sdl2_build_dir)
    rm_rf(sdl2_install_dir)
    extract(path, source_dir, "sdl2")

    mkdir_p(sdl2_build_dir)
    with cd(sdl2_build_dir):
        configuration = "Debug" if debug else "Release"
        cmake_args = cmake_args[:]
        cmake_args += [
            sdl2_source_dir,
            f"-DCMAKE_BUILD_TYPE={configuration}",
            f"-DCMAKE_INSTALL_PREFIX={cmake_path(sdl2_install_dir)}",
            "-DBUILD_SHARED_LIBS=OFF",
        ]
        if platform == "windows":
            cmake_args += [
                "-DSDL_FORCE_STATIC_VCRT=ON",
                "-DHAVE_LIBC=ON",
            ]
        elif platform == "macos":
            # システムでインストール済みかによって ON/OFF が切り替わってしまうため、
            # どの環境でも同じようにインストールされるようにするため全部 ON/OFF を明示的に指定する
            cmake_args += [
                "-DSDL_ATOMIC=OFF",
                "-DSDL_AUDIO=OFF",
                "-DSDL_VIDEO=ON",
                "-DSDL_RENDER=ON",
                "-DSDL_EVENTS=ON",
                "-DSDL_JOYSTICK=ON",
                "-DSDL_HAPTIC=ON",
                "-DSDL_POWER=ON",
                "-DSDL_THREADS=ON",
                "-DSDL_TIMERS=OFF",
                "-DSDL_FILE=OFF",
                "-DSDL_LOADSO=ON",
                "-DSDL_CPUINFO=OFF",
                "-DSDL_FILESYSTEM=OFF",
                "-DSDL_SENSOR=ON",
                "-DSDL_OPENGL=ON",
                "-DSDL_OPENGLES=ON",
                "-DSDL_RPI=OFF",
                "-DSDL_WAYLAND=OFF",
                "-DSDL_X11=OFF",
                "-DSDL_VULKAN=OFF",
                "-DSDL_VIVANTE=OFF",
                "-DSDL_COCOA=ON",
                "-DSDL_METAL=ON",
                "-DSDL_KMSDRM=OFF",
            ]
        elif platform == "linux":
            # システムでインストール済みかによって ON/OFF が切り替わってしまうため、
            # どの環境でも同じようにインストールされるようにするため全部 ON/OFF を明示的に指定する
            cmake_args += [
                "-DSDL_ATOMIC=OFF",
                "-DSDL_AUDIO=OFF",
                "-DSDL_VIDEO=ON",
                "-DSDL_RENDER=ON",
                "-DSDL_EVENTS=ON",
                "-DSDL_JOYSTICK=ON",
                "-DSDL_HAPTIC=ON",
                "-DSDL_POWER=ON",
                "-DSDL_THREADS=ON",
                "-DSDL_TIMERS=OFF",
                "-DSDL_FILE=OFF",
                "-DSDL_LOADSO=ON",
                "-DSDL_CPUINFO=OFF",
                "-DSDL_FILESYSTEM=OFF",
                "-DSDL_SENSOR=ON",
                "-DSDL_OPENGL=ON",
                "-DSDL_OPENGLES=ON",
                "-DSDL_RPI=OFF",
                "-DSDL_WAYLAND=OFF",
                "-DSDL_X11=ON",
                "-DSDL_X11_SHARED=OFF",
                "-DSDL_X11_XCURSOR=OFF",
                "-DSDL_X11_XDBE=OFF",
                "-DSDL_X11_XFIXES=OFF",
                "-DSDL_X11_XINERAMA=OFF",
                "-DSDL_X11_XINPUT=OFF",
                "-DSDL_X11_XRANDR=OFF",
                "-DSDL_X11_XSCRNSAVER=OFF",
                "-DSDL_X11_XSHAPE=OFF",
                "-DSDL_X11_XVM=OFF",
                "-DSDL_VULKAN=OFF",
                "-DSDL_VIVANTE=OFF",
                "-DSDL_COCOA=OFF",
                "-DSDL_METAL=OFF",
                "-DSDL_KMSDRM=OFF",
            ]
        cmd(["cmake"] + cmake_args)

        cmd(
            ["cmake", "--build", ".", "--config", configuration, f"-j{multiprocessing.cpu_count()}"]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_sdl3(
    version, source_dir, build_dir, install_dir, debug: bool, platform: str, cmake_args: List[str]
):
    url = f"https://github.com/libsdl-org/SDL/releases/download/release-{version}/SDL3-{version}.tar.gz"
    path = download(url, source_dir)
    sdl3_source_dir = os.path.join(source_dir, "sdl3")
    sdl3_build_dir = os.path.join(build_dir, "sdl3")
    sdl3_install_dir = os.path.join(install_dir, "sdl3")
    rm_rf(sdl3_source_dir)
    rm_rf(sdl3_build_dir)
    rm_rf(sdl3_install_dir)
    extract(path, source_dir, "sdl3")

    mkdir_p(sdl3_build_dir)
    with cd(sdl3_build_dir):
        configuration = "Debug" if debug else "Release"
        cmake_args = cmake_args[:]
        cmake_args += [
            sdl3_source_dir,
            f"-DCMAKE_BUILD_TYPE={configuration}",
            f"-DCMAKE_INSTALL_PREFIX={cmake_path(sdl3_install_dir)}",
            "-DBUILD_SHARED_LIBS=OFF",
            "-DSDL_STATIC=ON",
            "-DSDL_SHARED=OFF",
        ]
        if platform == "windows":
            cmake_args += [
                f"-DCMAKE_MSVC_RUNTIME_LIBRARY={'MultiThreaded' if not debug else 'MultiThreadedDebug'}",
                "-DSDL_AUDIO=OFF",
                "-DSDL_JOYSTICK=OFF",
                "-DSDL_HAPTIC=OFF",
                # GitHub Actions 上で gameinput.h が存在しないのに
                # なぜか  check_c_source_compiles() が成功してしまうので
                # HAVE_GAMEINPUT_H=0 で強制的に無効化する
                "-DHAVE_GAMEINPUT_H=0",
            ]
        elif platform == "macos":
            # どの環境でも同じようにインストールされるようにするため全部 ON/OFF を明示的に指定する
            cmake_args += [
                "-DSDL_AUDIO=OFF",
                "-DSDL_VIDEO=ON",
                "-DSDL_RENDER=ON",
                "-DSDL_HAPTIC=ON",
                "-DSDL_POWER=ON",
                "-DSDL_JOYSTICK=ON",
                "-DSDL_SENSOR=ON",
                "-DSDL_OPENGL=ON",
                "-DSDL_OPENGLES=ON",
                "-DSDL_METAL=ON",
                "-DSDL_VULKAN=OFF",
            ]
        elif platform == "linux":
            # どの環境でも同じようにインストールされるようにするため全部 ON/OFF を明示的に指定する
            cmake_args += [
                "-DSDL_AUDIO=OFF",
                "-DSDL_VIDEO=ON",
                "-DSDL_RENDER=ON",
                "-DSDL_HAPTIC=ON",
                "-DSDL_POWER=ON",
                "-DSDL_JOYSTICK=ON",
                "-DSDL_SENSOR=ON",
                "-DSDL_OPENGL=ON",
                "-DSDL_OPENGLES=ON",
                "-DSDL_X11=ON",
                "-DSDL_X11_SHARED=OFF",
                "-DSDL_X11_XCURSOR=OFF",
                "-DSDL_X11_XDBE=OFF",
                "-DSDL_X11_XFIXES=OFF",
                "-DSDL_X11_XINPUT=OFF",
                "-DSDL_X11_XRANDR=OFF",
                "-DSDL_X11_XSCRNSAVER=OFF",
                "-DSDL_X11_XSHAPE=OFF",
                "-DSDL_X11_XSYNC=OFF",
                "-DSDL_WAYLAND=OFF",
                "-DSDL_VULKAN=OFF",
                "-DSDL_KMSDRM=OFF",
                "-DSDL_RPI=OFF",
            ]
        cmd(["cmake"] + cmake_args)

        cmd(
            ["cmake", "--build", ".", "--config", configuration, f"-j{multiprocessing.cpu_count()}"]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_cli11(version, install_dir):
    cli11_install_dir = os.path.join(install_dir, "cli11")
    rm_rf(cli11_install_dir)
    cmd(
        [
            "git",
            "clone",
            "--branch",
            version,
            "--depth",
            "1",
            "https://github.com/CLIUtils/CLI11.git",
            cli11_install_dir,
        ]
    )


@versioned
def install_cuda_windows(version, source_dir, build_dir, install_dir):
    rm_rf(os.path.join(build_dir, "cuda"))
    rm_rf(os.path.join(install_dir, "cuda"))
    if version == "10.2.89-1":
        url = "http://developer.download.nvidia.com/compute/cuda/10.2/Prod/local_installers/cuda_10.2.89_441.22_win10.exe"  # noqa: E501
    elif version == "11.8.0-1":
        url = "https://developer.download.nvidia.com/compute/cuda/11.8.0/local_installers/cuda_11.8.0_522.06_windows.exe"  # noqa: E501
    else:
        raise Exception(f"Unknown CUDA version {version}")
    file = download(url, source_dir)

    mkdir_p(os.path.join(build_dir, "cuda"))
    mkdir_p(os.path.join(install_dir, "cuda"))
    with cd(os.path.join(build_dir, "cuda")):
        cmd(["7z", "x", file])
    copytree(
        os.path.join(build_dir, "cuda", "cuda_nvcc", "nvcc"), os.path.join(install_dir, "cuda")
    )
    copytree(
        os.path.join(build_dir, "cuda", "cuda_cudart", "cudart"), os.path.join(install_dir, "cuda")
    )


@versioned
def install_vpl(version, configuration, source_dir, build_dir, install_dir, cmake_args):
    vpl_source_dir = os.path.join(source_dir, "vpl")
    vpl_build_dir = os.path.join(build_dir, "vpl")
    vpl_install_dir = os.path.join(install_dir, "vpl")
    rm_rf(vpl_source_dir)
    rm_rf(vpl_build_dir)
    rm_rf(vpl_install_dir)
    git_clone_shallow("https://github.com/intel/libvpl.git", version, vpl_source_dir)

    mkdir_p(vpl_build_dir)
    with cd(vpl_build_dir):
        cmd(
            [
                "cmake",
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(vpl_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DBUILD_SHARED_LIBS=OFF",
                "-DBUILD_TOOLS=OFF",
                "-DBUILD_EXAMPLES=OFF",
                "-DBUILD_PREVIEW=OFF",
                "-DINSTALL_EXAMPLE_CODE=OFF",
                "-DBUILD_TOOLS_ONEVPL_EXPERIMENTAL=OFF",
                "-DUSE_MSVC_STATIC_RUNTIME=ON",
                vpl_source_dir,
                *cmake_args,
            ]
        )
        # 生成されたプロジェクトに対して静的ランタイムを使うように変更する
        vpl_path = os.path.join("libvpl", "VPL.vcxproj")
        if os.path.exists(vpl_path):
            replace_vcproj_static_runtime(vpl_path)

        cmd(
            ["cmake", "--build", ".", f"-j{multiprocessing.cpu_count()}", "--config", configuration]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_blend2d_official(
    version,
    configuration,
    source_dir,
    build_dir,
    install_dir,
    cmake_args,
    expected_sha256: Optional[str] = None,
):
    rm_rf(os.path.join(source_dir, "blend2d"))
    rm_rf(os.path.join(build_dir, "blend2d"))
    rm_rf(os.path.join(install_dir, "blend2d"))

    # 公式サイトに負荷をかけないための時雨堂によるミラー
    url = f"https://oss-mirrors.shiguredo.jp/blend2d-{version}.tar.gz"
    # Blend2d 公式
    # url = f"https://blend2d.com/download/blend2d-{version}.tar.gz"
    path = download(url, source_dir, expected_sha256=expected_sha256)
    extract(path, source_dir, "blend2d")
    _build_blend2d(
        configuration=configuration,
        source_dir=source_dir,
        build_dir=build_dir,
        install_dir=install_dir,
        cmake_args=cmake_args,
    )


@versioned
def install_blend2d(
    version,
    configuration,
    source_dir,
    build_dir,
    install_dir,
    blend2d_version,
    asmjit_version,
    cmake_args,
):
    rm_rf(os.path.join(source_dir, "blend2d"))
    rm_rf(os.path.join(build_dir, "blend2d"))
    rm_rf(os.path.join(install_dir, "blend2d"))

    git_clone_shallow(
        "https://github.com/blend2d/blend2d", blend2d_version, os.path.join(source_dir, "blend2d")
    )
    mkdir_p(os.path.join(source_dir, "blend2d", "3rdparty"))
    git_clone_shallow(
        "https://github.com/asmjit/asmjit",
        asmjit_version,
        os.path.join(source_dir, "blend2d", "3rdparty", "asmjit"),
    )
    _build_blend2d(
        configuration=configuration,
        source_dir=source_dir,
        build_dir=build_dir,
        install_dir=install_dir,
        cmake_args=cmake_args,
    )


def _build_blend2d(configuration, source_dir, build_dir, install_dir, cmake_args):
    mkdir_p(os.path.join(build_dir, "blend2d"))
    with cd(os.path.join(build_dir, "blend2d")):
        cmd(
            [
                "cmake",
                os.path.join(source_dir, "blend2d"),
                f"-DCMAKE_BUILD_TYPE={configuration}",
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(os.path.join(install_dir, 'blend2d'))}",
                "-DBLEND2D_STATIC=ON",
                *cmake_args,
            ]
        )
        # 生成されたプロジェクトに対して静的ランタイムを使うように変更する
        project_path = os.path.join(build_dir, "blend2d", "blend2d.vcxproj")
        if os.path.exists(project_path):
            replace_vcproj_static_runtime(project_path)

        cmd(
            [
                "cmake",
                "--build",
                ".",
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                configuration,
            ]
        )
        cmd(["cmake", "--build", ".", "--target", "install", "--config", configuration])


@versioned
def install_openh264(version, source_dir, install_dir, is_windows):
    rm_rf(os.path.join(source_dir, "openh264"))
    rm_rf(os.path.join(install_dir, "openh264"))
    git_clone_shallow(
        "https://github.com/cisco/openh264.git", version, os.path.join(source_dir, "openh264")
    )
    with cd(os.path.join(source_dir, "openh264")):
        if is_windows:
            # Windows は make が無いので手動でコピーする
            # install-headers:
            # 	mkdir -p $(DESTDIR)$(PREFIX)/include/wels
            # 	install -m 644 $(SRC_PATH)/codec/api/wels/codec*.h $(DESTDIR)$(PREFIX)/include/wels
            mkdir_p(os.path.join(install_dir, "openh264", "include", "wels"))
            with cd(os.path.join("codec", "api", "wels")):
                for file in glob.glob("codec*.h"):
                    shutil.copyfile(
                        file, os.path.join(install_dir, "openh264", "include", "wels", file)
                    )
        else:
            cmd(["make", f"PREFIX={os.path.join(install_dir, 'openh264')}", "install-headers"])


@versioned
def install_yaml(version, source_dir, build_dir, install_dir, cmake_args):
    rm_rf(os.path.join(source_dir, "yaml"))
    rm_rf(os.path.join(install_dir, "yaml"))
    rm_rf(os.path.join(build_dir, "yaml"))
    git_clone_shallow(
        "https://github.com/jbeder/yaml-cpp.git", version, os.path.join(source_dir, "yaml")
    )

    mkdir_p(os.path.join(build_dir, "yaml"))
    with cd(os.path.join(build_dir, "yaml")):
        cmd(
            [
                "cmake",
                os.path.join(source_dir, "yaml"),
                "-DCMAKE_BUILD_TYPE=Release",
                f"-DCMAKE_INSTALL_PREFIX={install_dir}/yaml",
                "-DYAML_CPP_BUILD_TESTS=OFF",
                "-DYAML_CPP_BUILD_TOOLS=OFF",
                *cmake_args,
            ]
        )
        cmd(["cmake", "--build", ".", f"-j{multiprocessing.cpu_count()}"])
        cmd(["cmake", "--build", ".", "--target", "install"])


@versioned
def install_catch2(version, source_dir, build_dir, install_dir, configuration, cmake_args):
    rm_rf(os.path.join(source_dir, "catch2"))
    rm_rf(os.path.join(install_dir, "catch2"))
    rm_rf(os.path.join(build_dir, "catch2"))
    git_clone_shallow(
        "https://github.com/catchorg/Catch2.git", version, os.path.join(source_dir, "catch2")
    )

    mkdir_p(os.path.join(build_dir, "catch2"))
    with cd(os.path.join(build_dir, "catch2")):
        cmd(
            [
                "cmake",
                os.path.join(source_dir, "catch2"),
                f"-DCMAKE_BUILD_TYPE={configuration}",
                f"-DCMAKE_INSTALL_PREFIX={install_dir}/catch2",
                "-DCATCH_BUILD_TESTING=OFF",
                *cmake_args,
            ]
        )
        # 生成されたプロジェクトに対して静的ランタイムを使うように変更する
        project_path = os.path.join("src", "Catch2.vcxproj")
        if os.path.exists(project_path):
            replace_vcproj_static_runtime(project_path)
        project_path = os.path.join("src", "Catch2WithMain.vcxproj")
        if os.path.exists(project_path):
            replace_vcproj_static_runtime(project_path)
        cmd(
            ["cmake", "--build", ".", "--config", configuration, f"-j{multiprocessing.cpu_count()}"]
        )
        cmd(["cmake", "--build", ".", "--config", configuration, "--target", "install"])


@versioned
def install_protobuf(version, source_dir, install_dir, platform: str):
    # platform:
    # - linux-aarch_64
    # - linux-ppcle_64
    # - linux-s390_64
    # - linux-x86_32
    # - linux-x86_64
    # - osx-aarch_64
    # - osx-universal_binary
    # - osx-x86_64
    # - win32
    # - win64
    url = f"https://github.com/protocolbuffers/protobuf/releases/download/v{version}/protoc-{version}-{platform}.zip"
    path = download(url, source_dir)
    rm_rf(os.path.join(install_dir, "protobuf"))
    extract(path, install_dir, "protobuf")
    # なぜか実行属性が消えてるので入れてやる
    for file in os.scandir(os.path.join(install_dir, "protobuf", "bin")):
        if file.is_file():
            os.chmod(file.path, file.stat().st_mode | stat.S_IXUSR)


@versioned
def install_protoc_gen_jsonif(version, source_dir, install_dir, platform: str):
    # platform:
    # - darwin-amd64
    # - darwin-arm64
    # - linux-amd64
    # - windows-amd64
    url = f"https://github.com/melpon/protoc-gen-jsonif/releases/download/{version}/protoc-gen-jsonif.tar.gz"
    rm_rf(os.path.join(source_dir, "protoc-gen-jsonif.tar.gz"))
    path = download(url, source_dir)
    jsonif_install_dir = os.path.join(install_dir, "protoc-gen-jsonif")
    rm_rf(jsonif_install_dir)
    extract(path, install_dir, "protoc-gen-jsonif")
    # 自分の環境のバイナリを <install-path>/bin に配置する
    shutil.copytree(
        os.path.join(jsonif_install_dir, *platform.split("-")),
        os.path.join(jsonif_install_dir, "bin"),
    )
    # なぜか実行属性が消えてるので入れてやる
    for file in os.scandir(os.path.join(jsonif_install_dir, "bin")):
        if file.is_file():
            os.chmod(file.path, file.stat().st_mode | stat.S_IXUSR)


# iOS, Android などのクロスコンパイル環境で実行可能ファイルを生成しようとしてエラーになるのを防止するパッチ
#
# v1.64.1 をベースにパッチを当てている
#
# MEMO: gRPC の、submodule を含めて全ての diff を取得するコマンド
#   git --no-pager diff --ignore-submodules && git submodule foreach --recursive 'git --no-pager diff --ignore-submodules --src-prefix a/$path/ --dst-prefix b/$path/' | grep -v '^Entering'
GRPC_PATCH_NO_EXECUTABLE = r"""
diff --git a/third_party/boringssl-with-bazel/CMakeLists.txt b/third_party/boringssl-with-bazel/CMakeLists.txt
index 6464e200f..c7bc417a1 100644
--- a/third_party/boringssl-with-bazel/CMakeLists.txt
+++ b/third_party/boringssl-with-bazel/CMakeLists.txt
@@ -543,30 +543,6 @@ add_library(
 
 target_link_libraries(ssl crypto)
 
-add_executable(
-  bssl
-
-  src/tool/args.cc
-  src/tool/ciphers.cc
-  src/tool/client.cc
-  src/tool/const.cc
-  src/tool/digest.cc
-  src/tool/fd.cc
-  src/tool/file.cc
-  src/tool/generate_ech.cc
-  src/tool/generate_ed25519.cc
-  src/tool/genrsa.cc
-  src/tool/pkcs12.cc
-  src/tool/rand.cc
-  src/tool/server.cc
-  src/tool/sign.cc
-  src/tool/speed.cc
-  src/tool/tool.cc
-  src/tool/transport_common.cc
-)
-
-target_link_libraries(bssl ssl crypto)
-
 if(NOT CMAKE_SYSTEM_NAME STREQUAL "Android")
   find_package(Threads REQUIRED)
   target_link_libraries(crypto Threads::Threads)
diff --git a/third_party/zlib/CMakeLists.txt b/third_party/zlib/CMakeLists.txt
index 7f1b69f..bcf5577 100644
--- a/third_party/zlib/CMakeLists.txt
+++ b/third_party/zlib/CMakeLists.txt
@@ -147,10 +147,7 @@ if(MINGW)
     set(ZLIB_DLL_SRCS ${CMAKE_CURRENT_BINARY_DIR}/zlib1rc.obj)
 endif(MINGW)
 
-add_library(zlib SHARED ${ZLIB_SRCS} ${ZLIB_DLL_SRCS} ${ZLIB_PUBLIC_HDRS} ${ZLIB_PRIVATE_HDRS})
 add_library(zlibstatic STATIC ${ZLIB_SRCS} ${ZLIB_PUBLIC_HDRS} ${ZLIB_PRIVATE_HDRS})
-set_target_properties(zlib PROPERTIES DEFINE_SYMBOL ZLIB_DLL)
-set_target_properties(zlib PROPERTIES SOVERSION 1)
 
 if(NOT CYGWIN)
     # This property causes shared libraries on Linux to have the full version
@@ -160,22 +157,16 @@ if(NOT CYGWIN)
     #
     # This has no effect with MSVC, on that platform the version info for
     # the DLL comes from the resource file win32/zlib1.rc
-    set_target_properties(zlib PROPERTIES VERSION ${ZLIB_FULL_VERSION})
 endif()
 
 if(UNIX)
     # On unix-like platforms the library is almost always called libz
-   set_target_properties(zlib zlibstatic PROPERTIES OUTPUT_NAME z)
-   if(NOT APPLE)
-     set_target_properties(zlib PROPERTIES LINK_FLAGS "-Wl,--version-script,\"${CMAKE_CURRENT_SOURCE_DIR}/zlib.map\"")
-   endif()
 elseif(BUILD_SHARED_LIBS AND WIN32)
     # Creates zlib1.dll when building shared library version
-    set_target_properties(zlib PROPERTIES SUFFIX "1.dll")
 endif()
 
 if(NOT SKIP_INSTALL_LIBRARIES AND NOT SKIP_INSTALL_ALL )
-    install(TARGETS zlib zlibstatic
+    install(TARGETS zlibstatic
         RUNTIME DESTINATION "${INSTALL_BIN_DIR}"
         ARCHIVE DESTINATION "${INSTALL_LIB_DIR}"
         LIBRARY DESTINATION "${INSTALL_LIB_DIR}" )
@@ -193,21 +184,3 @@ endif()
 #============================================================================
 # Example binaries
 #============================================================================
-
-add_executable(example test/example.c)
-target_link_libraries(example zlib)
-add_test(example example)
-
-add_executable(minigzip test/minigzip.c)
-target_link_libraries(minigzip zlib)
-
-if(HAVE_OFF64_T)
-    add_executable(example64 test/example.c)
-    target_link_libraries(example64 zlib)
-    set_target_properties(example64 PROPERTIES COMPILE_FLAGS "-D_FILE_OFFSET_BITS=64")
-    add_test(example64 example64)
-
-    add_executable(minigzip64 test/minigzip.c)
-    target_link_libraries(minigzip64 zlib)
-    set_target_properties(minigzip64 PROPERTIES COMPILE_FLAGS "-D_FILE_OFFSET_BITS=64")
-endif()
"""


@versioned
def install_grpc(
    version,
    source_dir,
    build_dir,
    install_dir,
    debug: bool,
    cmake_args: List[str],
    cmake_build_args: List[str] = [],
):
    grpc_source_dir = os.path.join(source_dir, "grpc")
    grpc_build_dir = os.path.join(build_dir, "grpc")
    grpc_install_dir = os.path.join(install_dir, "grpc")
    rm_rf(grpc_source_dir)
    rm_rf(grpc_build_dir)
    rm_rf(grpc_install_dir)
    git_clone_shallow("https://github.com/grpc/grpc.git", version, grpc_source_dir, submodule=True)
    apply_patch_text(GRPC_PATCH_NO_EXECUTABLE, grpc_source_dir, 1)
    mkdir_p(grpc_build_dir)
    with cd(grpc_build_dir):
        configuration = "Debug" if debug else "Release"
        cmd(
            [
                "cmake",
                grpc_source_dir,
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(grpc_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                *cmake_args,
            ]
        )
        cmd(
            [
                "cmake",
                "--build",
                ".",
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                configuration,
                *cmake_build_args,
            ]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_ggrpc(version, install_dir):
    ggrpc_install_dir = os.path.join(install_dir, "ggrpc")
    rm_rf(ggrpc_install_dir)
    git_clone_shallow("https://github.com/melpon/ggrpc.git", version, ggrpc_install_dir)


@versioned
def install_spdlog(version, install_dir):
    spdlog_install_dir = os.path.join(install_dir, "spdlog")
    rm_rf(spdlog_install_dir)
    git_clone_shallow("https://github.com/gabime/spdlog.git", version, spdlog_install_dir)


BORINGSSL_PATCH_NO_BSSL = r"""
diff --git a/CMakeLists.txt b/CMakeLists.txt
index 38d63db..b97b175 100644
--- a/CMakeLists.txt
+++ b/CMakeLists.txt
@@ -795,7 +795,7 @@ endif()
 
 if(INSTALL_ENABLED)
   install(TARGETS crypto ssl EXPORT OpenSSLTargets)
-  install(TARGETS bssl)
+  # install(TARGETS bssl)
   install(DIRECTORY include/ DESTINATION ${CMAKE_INSTALL_INCLUDEDIR})
   install(EXPORT OpenSSLTargets
           FILE OpenSSLTargets.cmake
"""


@versioned
def install_boringssl(
    version,
    source_dir: str,
    build_dir: str,
    install_dir: str,
    configuration: str,
    cmake_args: List[str],
):
    boringssl_source_dir = os.path.join(source_dir, "boringssl")
    boringssl_build_dir = os.path.join(build_dir, "boringssl")
    boringssl_install_dir = os.path.join(install_dir, "boringssl")
    rm_rf(boringssl_source_dir)
    rm_rf(boringssl_build_dir)
    rm_rf(boringssl_install_dir)
    git_clone_shallow("https://boringssl.googlesource.com/boringssl", version, boringssl_source_dir)
    apply_patch_text(BORINGSSL_PATCH_NO_BSSL, boringssl_source_dir, 1)
    mkdir_p(boringssl_build_dir)
    with cd(boringssl_build_dir):
        cmd(
            [
                "cmake",
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(boringssl_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DBUILD_SHARED_LIBS=OFF",
                boringssl_source_dir,
                *cmake_args,
            ]
        )
        cmd(
            ["cmake", "--build", ".", f"-j{multiprocessing.cpu_count()}", "--config", configuration]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_opus(
    version, source_dir, build_dir, install_dir, configuration: str, cmake_args: List[str]
):
    opus_source_dir = os.path.join(source_dir, "opus")
    opus_build_dir = os.path.join(build_dir, "opus")
    opus_install_dir = os.path.join(install_dir, "opus")
    rm_rf(opus_source_dir)
    rm_rf(opus_build_dir)
    rm_rf(opus_install_dir)
    git_clone_shallow("https://gitlab.xiph.org/xiph/opus", version, opus_source_dir)
    mkdir_p(opus_build_dir)
    with cd(opus_build_dir):
        cmd(
            [
                "cmake",
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(opus_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DCMAKE_POSITION_INDEPENDENT_CODE=ON",
                "-DOPUS_BUILD_SHARED_LIBRARY=OFF",
                "-DOPUS_BUILD_TESTING=OFF",
                "-DOPUS_BUILD_PROGRAMS=OFF",
                "-DOPUS_STATIC_RUNTIME=ON",
                opus_source_dir,
                *cmake_args,
            ]
        )
        cmd(
            ["cmake", "--build", ".", f"-j{multiprocessing.cpu_count()}", "--config", configuration]
        )
        cmd(["cmake", "--install", ".", "--config", configuration])


@versioned
def install_nasm(version, source_dir, install_dir, platform: str):
    if platform not in ("macosx", "win32", "win64"):
        raise Exception(f"Unsupported platform: {platform}")
    url = f"https://www.nasm.us/pub/nasm/releasebuilds/{version}/{platform}/nasm-{version}-{platform}.zip"
    path = download(url, source_dir)
    nasm_install_dir = os.path.join(install_dir, "nasm")
    rm_rf(nasm_install_dir)
    extract(path, install_dir, "nasm")


@versioned
def install_ninja(version, source_dir, install_dir, platform):
    if platform not in ("win", "winarm64", "linux-aarch64", "linux", "mac"):
        raise Exception(f"Unsupported platform: {platform}")
    url = f"https://github.com/ninja-build/ninja/releases/download/{version}/ninja-{platform}.zip"
    path = download(url, source_dir)
    ninja_install_dir = os.path.join(install_dir, "ninja")
    rm_rf(ninja_install_dir)
    extract(path, install_dir, "ninja")


@versioned
def install_vswhere(version, install_dir):
    url = f"https://github.com/microsoft/vswhere/releases/download/{version}/vswhere.exe"
    vswhere_install_dir = os.path.join(install_dir, "vswhere")
    rm_rf(vswhere_install_dir)
    mkdir_p(vswhere_install_dir)
    download(url, vswhere_install_dir)


@versioned
def install_amf(version, install_dir):
    # AMF はライブラリというよりは便利関数＋サンプルコードという感じなので
    # ここではソースだけダウンロードをして、必要に応じて利用者側でファイルを参照してインクルード/コンパイルする
    amf_install_dir = os.path.join(install_dir, "amf")
    rm_rf(amf_install_dir)
    git_clone_shallow(
        "https://github.com/GPUOpen-LibrariesAndSDKs/AMF.git", version, amf_install_dir
    )


@versioned
def install_mbedtls(version, source_dir, build_dir, install_dir, debug, cmake_args):
    rm_rf(os.path.join(source_dir, "mbedtls"))
    rm_rf(os.path.join(build_dir, "mbedtls"))
    rm_rf(os.path.join(install_dir, "mbedtls"))
    git_clone_shallow(
        "https://github.com/Mbed-TLS/mbedtls.git",
        version,
        os.path.join(source_dir, "mbedtls"),
    )
    with cd(os.path.join(source_dir, "mbedtls")):
        configuration = "Debug" if debug else "Release"
        cmd(["python3", "scripts/config.py", "set", "MBEDTLS_SSL_DTLS_SRTP"])
        cmd(
            [
                "cmake",
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(os.path.join(install_dir, 'mbedtls'))}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DCMAKE_POSITION_INDEPENDENT_CODE=ON",
                "-B",
                os.path.join(build_dir, "mbedtls"),
            ]
            + cmake_args
        )
        cmd(
            [
                "cmake",
                "--build",
                os.path.join(build_dir, "mbedtls"),
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                "Release",
            ]
        )
        cmd(["cmake", "--install", os.path.join(build_dir, "mbedtls")])


@versioned
def install_libjpeg_turbo(version, source_dir, build_dir, install_dir, configuration, cmake_args):
    libjpeg_source_dir = os.path.join(source_dir, "libjpeg-turbo")
    libjpeg_build_dir = os.path.join(build_dir, "libjpeg-turbo")
    libjpeg_install_dir = os.path.join(install_dir, "libjpeg-turbo")
    rm_rf(libjpeg_source_dir)
    rm_rf(libjpeg_build_dir)
    rm_rf(libjpeg_install_dir)
    git_clone_shallow(
        "https://github.com/libjpeg-turbo/libjpeg-turbo.git",
        version,
        libjpeg_source_dir,
    )
    mkdir_p(libjpeg_build_dir)
    with cd(libjpeg_build_dir):
        cmd(
            [
                "cmake",
                libjpeg_source_dir,
                f"-DCMAKE_INSTALL_PREFIX={libjpeg_install_dir}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DENABLE_SHARED=FALSE",
                "-DENABLE_STATIC=TRUE",
                "-DCMAKE_BUILD_TYPE=Release",
                "-DCMAKE_POSITION_INDEPENDENT_CODE=ON",
            ]
            + cmake_args
        )
        cmd(
            [
                "cmake",
                "--build",
                libjpeg_build_dir,
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                "Release",
            ]
        )
        cmd(["cmake", "--install", libjpeg_build_dir])


@versioned
def install_libyuv(
    version, source_dir, build_dir, install_dir, libjpeg_turbo_dir, configuration, cmake_args
):
    libyuv_source_dir = os.path.join(source_dir, "libyuv")
    libyuv_build_dir = os.path.join(build_dir, "libyuv")
    libyuv_install_dir = os.path.join(install_dir, "libyuv")
    rm_rf(libyuv_source_dir)
    rm_rf(libyuv_build_dir)
    rm_rf(libyuv_install_dir)
    git_clone_shallow(
        "https://chromium.googlesource.com/libyuv/libyuv",
        version,
        libyuv_source_dir,
    )
    mkdir_p(libyuv_build_dir)
    with cd(libyuv_build_dir):
        cmd(
            [
                "cmake",
                libyuv_source_dir,
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(libyuv_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                f"-DCMAKE_PREFIX_PATH={cmake_path(libjpeg_turbo_dir)}",
                *cmake_args,
            ]
        )
        cmd(
            [
                "cmake",
                "--build",
                libyuv_build_dir,
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                "Release",
            ]
        )
        cmd(["cmake", "--install", libyuv_build_dir])


@versioned
def install_aom(version, source_dir, build_dir, install_dir, configuration, cmake_args):
    aom_source_dir = os.path.join(source_dir, "aom")
    aom_build_dir = os.path.join(build_dir, "aom")
    aom_install_dir = os.path.join(install_dir, "aom")
    rm_rf(aom_source_dir)
    rm_rf(aom_build_dir)
    rm_rf(aom_install_dir)
    git_clone_shallow(
        "https://aomedia.googlesource.com/aom",
        version,
        aom_source_dir,
    )
    with cd(aom_source_dir):
        cmd(
            [
                "cmake",
                "-B",
                aom_build_dir,
                f"-DCMAKE_INSTALL_PREFIX={cmake_path(aom_install_dir)}",
                f"-DCMAKE_BUILD_TYPE={configuration}",
                "-DENABLE_DOCS=OFF",
                "-DENABLE_TESTS=OFF",
                *cmake_args,
            ]
        )
        cmd(
            [
                "cmake",
                "--build",
                aom_build_dir,
                f"-j{multiprocessing.cpu_count()}",
                "--config",
                "Release",
            ]
        )
        cmd(["cmake", "--install", aom_build_dir])


class PlatformTarget(object):
    def __init__(self, os, osver, arch, extra=None):
        self.os = os
        self.osver = osver
        self.arch = arch
        self.extra = extra

    @property
    def package_name(self):
        if self.os == "windows":
            return f"windows_{self.arch}"
        if self.os == "macos":
            return f"macos_{self.arch}"
        if self.os == "ubuntu":
            return f"ubuntu-{self.osver}_{self.arch}"
        if self.os == "ios":
            return "ios"
        if self.os == "android":
            return "android"
        if self.os == "raspberry-pi-os":
            return f"raspberry-pi-os_{self.arch}"
        if self.os == "jetson":
            if self.extra is None:
                ubuntu_version = "ubuntu-20.04"
            else:
                ubuntu_version = self.extra
            if self.osver is None:
                return f"{ubuntu_version}_armv8_jetson"
            return f"{ubuntu_version}_armv8_jetson_{self.osver}"
        raise Exception("error")


def get_windows_osver():
    osver = platform.release()
    # Windows 以外の環境だと reportAttributeAccessIssue を報告されてしまうので ignore する
    with winreg.OpenKeyEx(  # type: ignore[reportAttributeAccessIssue]
        winreg.HKEY_LOCAL_MACHINE,  # type: ignore[reportAttributeAccessIssue]
        "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",
    ) as key:
        return osver + "." + winreg.QueryValueEx(key, "ReleaseId")[0]  # type: ignore[reportAttributeAccessIssue]


def get_macos_osver():
    platform.mac_ver()[0]


def get_build_platform() -> PlatformTarget:
    os = platform.system()
    if os == "Windows":
        os = "windows"
        osver = get_windows_osver()
    elif os == "Darwin":
        os = "macos"
        osver = get_macos_osver()
    elif os == "Linux":
        release = read_version_file("/etc/os-release")
        os = release["NAME"]
        if os == "Ubuntu":
            os = "ubuntu"
            osver = release["VERSION_ID"]
        else:
            raise Exception(f"OS {os} not supported")
        pass
    else:
        raise Exception(f"OS {os} not supported")

    arch = platform.machine()
    if arch in ("AMD64", "x86_64"):
        arch = "x86_64"
    elif arch in ("aarch64", "arm64"):
        if os == "ubuntu":
            arch = "armv8"
        else:
            arch = "arm64"
    else:
        raise Exception(f"Arch {arch} not supported")

    return PlatformTarget(os, osver, arch)


def get_clang_version(clang):
    version_str = cmdcap([clang, "--version"])

    # version_str は以下のような文字列になっているので、ここからバージョンを取る
    #
    # clang version 16.0.0 (...)
    # Target: x86_64-unknown-linux-gnu
    # Thread model: posix
    # InstalledDir: /path/to/clang/bin
    #
    # Android 版だと以下のような文字列になっている
    #
    # Android (8490178, based on r450784d) clang version 14.0.6 (...)
    # Target: aarch64-unknown-linux-android29
    # Thread model: posix
    # InstalledDir: /path/to/android-ndk/toolchains/llvm/prebuilt/linux-x86_64/bin

    # clang version の次の文字列を取る
    xs = version_str.split("\n")[0].split(" ")
    for i in range(2, len(xs)):
        if xs[i - 2] == "clang" and xs[i - 1] == "version":
            return xs[i]

    raise Exception("Failed to get clang version")


def fix_clang_version(clang_dir, clang_version):
    # <clang_dir>/lib/clang/<clang_version>/include または
    # <clang_dir>/lib64/clang/<clang_version>/include が存在するか調べて、
    # 存在しない場合は clang_version を調節して、存在するバージョンに変換する
    #
    # <clang_dir>/lib/clang/16.0.0/include になっている場合と
    # <clang_dir>/lib/clang/16/include になっている場合があるため
    paths = [os.path.join(clang_dir, "lib", "clang"), os.path.join(clang_dir, "lib64", "clang")]
    exists = any(map(lambda x: os.path.exists(os.path.join(x, clang_version, "include")), paths))
    if exists:
        return clang_version

    fixed_clang_version = clang_version.split(".")[0]
    exists = any(
        map(lambda x: os.path.exists(os.path.join(x, fixed_clang_version, "include")), paths)
    )
    if exists:
        return fixed_clang_version

    raise Exception(
        f"Failed to fix clang version: clang_dir={clang_dir} clang_version={clang_version}"
    )


class Platform(object):
    def _check(self, flag, error_message):
        if not flag:
            raise Exception(error_message)

    def _check_platform_target(self, p: PlatformTarget):
        if p.os == "raspberry-pi-os":
            self._check(
                p.arch in ("armv6", "armv7", "armv8"),
                f"Architecture {p.arch} is not supported for {p.os}. Supported: armv6, armv7, armv8",
            )
        elif p.os == "jetson":
            self._check(
                p.arch == "armv8",
                f"Architecture {p.arch} is not supported for {p.os}. Only armv8 is supported",
            )
        elif p.os in ("ios", "android"):
            self._check(p.arch is None, f"Architecture should be None for {p.os}, but got {p.arch}")
        elif p.os == "ubuntu":
            self._check(
                p.arch in ("x86_64", "armv8"),
                f"Architecture {p.arch} is not supported for {p.os}. Supported: x86_64, armv8",
            )
        else:
            self._check(
                p.arch in ("x86_64", "arm64", "hololens2"),
                f"Architecture {p.arch} is not supported for {p.os}. Supported: x86_64, arm64, hololens2",
            )

    def __init__(self, target_os, target_osver, target_arch, target_extra=None):
        build = get_build_platform()
        target = PlatformTarget(target_os, target_osver, target_arch, target_extra)

        self._check_platform_target(build)
        self._check_platform_target(target)

        if target.os == "windows":
            self._check(
                target.arch in ("x86_64", "arm64", "hololens2"),
                f"Target architecture {target.arch} is not supported for Windows",
            )
            self._check(
                build.os == "windows",
                f"Windows target requires Windows build platform, but got {build.os}",
            )
            self._check(
                build.arch == "x86_64",
                f"Windows build requires x86_64 architecture, but got {build.arch}",
            )
        if target.os == "macos":
            self._check(
                build.os == "macos",
                f"macOS target requires macOS build platform, but got {build.os}",
            )
            self._check(
                build.arch in ("x86_64", "arm64"),
                f"macOS build requires x86_64 or arm64, but got {build.arch}",
            )
        if target.os == "ios":
            self._check(
                build.os == "macos", f"iOS target requires macOS build platform, but got {build.os}"
            )
            self._check(
                build.arch in ("x86_64", "arm64"),
                f"iOS build requires x86_64 or arm64, but got {build.arch}",
            )
        if target.os == "android":
            self._check(
                build.os in ("ubuntu", "macos"),
                f"Android target requires Ubuntu or macOS build platform, but got {build.os}",
            )
            if build.os == "ubuntu":
                self._check(
                    build.arch == "x86_64",
                    f"Android build on Ubuntu requires x86_64, but got {build.arch}",
                )
            elif build.os == "macos":
                self._check(
                    build.arch in ("x86_64", "arm64"),
                    f"Android build on macOS requires x86_64 or arm64, but got {build.arch}",
                )
        if target.os == "ubuntu":
            self._check(
                build.os == "ubuntu",
                f"Ubuntu target requires Ubuntu build platform, but got {build.os}",
            )
            self._check(
                build.arch in ("x86_64", "armv8"),
                f"Ubuntu build requires x86_64 or armv8, but got {build.arch}",
            )
            if build.arch == target.arch:
                self._check(
                    build.osver == target.osver,
                    f"When building for the same architecture ({build.arch}), OS versions must match. "
                    f"Build OS: Ubuntu {build.osver}, Target OS: Ubuntu {target.osver}. "
                    f"This typically happens when GitHub Actions runner OS version doesn't match the target.",
                )
        if target.os == "raspberry-pi-os":
            self._check(
                build.os == "ubuntu",
                f"Raspberry Pi OS target requires Ubuntu build platform, but got {build.os}",
            )
            self._check(
                build.arch == "x86_64",
                f"Raspberry Pi OS build requires x86_64, but got {build.arch}",
            )
        if target.os == "jetson":
            self._check(
                build.os == "ubuntu",
                f"Jetson target requires Ubuntu build platform, but got {build.os}",
            )
            self._check(
                build.arch == "x86_64", f"Jetson build requires x86_64, but got {build.arch}"
            )

        self.build = build
        self.target = target


def get_webrtc_platform(platform: Platform) -> str:
    # WebRTC
    if platform.target.os == "windows":
        return f"windows_{platform.target.arch}"
    elif platform.target.os == "macos":
        return f"macos_{platform.target.arch}"
    elif platform.target.os == "ios":
        return "ios"
    elif platform.target.os == "android":
        return "android"
    elif platform.target.os == "ubuntu":
        return f"ubuntu-{platform.target.osver}_{platform.target.arch}"
    elif platform.target.os == "raspberry-pi-os":
        return f"raspberry-pi-os_{platform.target.arch}"
    elif platform.target.os == "jetson":
        if platform.target.extra is None:
            return "ubuntu-20.04_armv8"
        else:
            return f"{platform.target.extra}_armv8"
    else:
        raise Exception(f"Unknown platform {platform.target.os}")


# 内部で os.path.abspath() を利用しており、 os.path.abspath() はカレントディレクトリに依存するため、
# この関数を利用する場合は ArgumentParser.parse_args() 実行前にカレントディレクトリを変更してはならない
#
# また、 --sora-args の指定には `--sora-args='--test'` のように `=` を使う必要がある
# `--sora-args '--test'` のようにスペースを使うと、ハイフンから始まるオプションが正しく解釈されない
def add_sora_arguments(parser):
    parser.add_argument(
        "--local-sora-cpp-sdk-dir",
        type=os.path.abspath,
        default=None,
        help="Refer to local Sora C++ SDK. "
        "When this option is specified, Sora C++ SDK will also be built.",
    )
    parser.add_argument(
        "--local-sora-cpp-sdk-args",
        type=shlex.split,
        default=[],
        help="Options for building local Sora C++ SDK when `--local-sora-cpp-sdk-dir` is specified.",
    )


# add_sora_arguments と同様の注意点があるので注意すること
def add_webrtc_build_arguments(parser):
    parser.add_argument(
        "--local-webrtc-build-dir",
        type=os.path.abspath,
        default=None,
        help="Refer to local webrtc-build. "
        "When this option is specified, webrtc-build will also be built.",
    )
    parser.add_argument(
        "--local-webrtc-build-args",
        type=shlex.split,
        default=[],
        help="Options for building local webrtc-build when `--local-webrtc-build-dir` is specified.",
    )
